# pylint: disable=attribute-defined-outside-init, len-as-condition, unused-argument

def load_landmarks(filename):
    import re
    from baiji import s3
    from baiji.serialization import json, pickle, yaml
    from lace.serialization import lmrk, meshlab_pickedpoints

    if not s3.exists(filename):
        raise ValueError("Landmark file %s not found" % filename)
    if re.search(".ya{0,1}ml$", filename):
        return yaml.load(filename)
    elif re.search(".json$", filename):
        return json.load(filename)
    elif re.search(".pkl$", filename):
        return pickle.load(filename)
    elif re.search(".lmrk$", filename):
        return lmrk.load(filename)
    elif re.search(".pp$", filename):
        return meshlab_pickedpoints.load(filename)
    else:
        raise ValueError("Landmark file %s is of unknown format" % filename)

class MeshMixin(object):
    '''
    landm: a dictionary of indices
    landm_xyz: landmark locations; these are basically v[landm]
    landm_regressors: regressors; replaces landm, predicts landm_xyz
    '''

    @property
    def landm_xyz(self, ordering=None):
        if ordering is None:
            ordering = self.landm_names

        if len(ordering) == 0: # there are no landmarks
            return {}

        landmark_vertex_locations = (self.landm_xyz_linear_transform(ordering) * self.v.flatten()).reshape(-1, 3)
        return dict(zip(ordering, landmark_vertex_locations))

    def landm_xyz_linear_transform(self, ordering=None):
        import numpy as np
        from blmath.numerics.matlab import col, sparse
        if ordering is None:
            ordering = self.landm_names
        # construct a sparse matrix that converts between the landmark pts and all vertices, with height (# landmarks * 3) and width (# vertices * 3)
        if self.landm_regressors is not None:
            landmark_coefficients = np.hstack([self.landm_regressors[name][1] for name in ordering])
            landmark_indices = np.hstack([self.landm_regressors[name][0] for name in ordering])
            column_indices = np.hstack([col(3*landmark_indices + i) for i in range(3)]).flatten()
            row_indices = np.hstack([[3*index, 3*index + 1, 3*index + 2]*len(self.landm_regressors[ordering[index]][0]) for index in np.arange(len(ordering))])
            values = np.hstack([col(landmark_coefficients) for i in range(3)]).flatten()
            return sparse(row_indices, column_indices, values, 3*len(ordering), 3*self.v.shape[0])
        elif hasattr(self, 'landm') and len(self.landm) > 0:
            landmark_indices = np.array([self.landm[name] for name in ordering])
            column_indices = np.hstack(([col(3*landmark_indices + i) for i in range(3)])).flatten()
            row_indices = np.arange(3*len(ordering))
            return sparse(row_indices, column_indices, np.ones(len(column_indices)), 3*len(ordering), 3*self.v.shape[0])
        else:
            return np.zeros((0, 0))

    @landm_xyz.setter
    def landm_xyz(self, val):
        '''
        Since landm can handle all possibilities, we just delegate to that
        '''
        self.landm = val

    @property
    def landm_regressors(self):
        if hasattr(self, '_landm_regressors'):
            return self._landm_regressors
        else:
            return None

    @landm_regressors.setter
    def landm_regressors(self, val):
        if val is None or len(val) == 0:
            self._landm_regressors = None
        else:
            self._landm_regressors = val
            self.landm = None

    @property
    def landm(self):
        if hasattr(self, '_landm'):
            return self._landm
        else:
            return None
    @landm.setter
    def landm(self, val):
        '''
        Sets landmarks given any of:
         - ppfile
         - ldmk file
         - dict of {name:inds} (i.e. mesh.landm)
         - dict of {name:xyz} (i.e. mesh.landm_xyz)
         - Nx1 array or list of ints (treated as landm, given sequential integers as names)
         - Nx3 array or list of floats (treated as landm_xyz, given sequential integers as names)
         - pkl, json, yaml file containing either of the above dicts or arrays
        '''
        import numpy as np

        if val is None:
            self._landm = None
            self._raw_landmarks = None
        elif isinstance(val, basestring):
            self.landm = load_landmarks(val)
        else:
            if not hasattr(val, 'keys'):
                val = {str(ii): v for ii, v in enumerate(val)}
            landm = {}
            landm_xyz = {}
            filtered_landmarks = []
            for k, v in val.iteritems():
                if isinstance(v, (int, long)):
                    landm[k] = v
                elif len(v) == 3:
                    if np.all(v == [0.0, 0.0, 0.0]):
                        filtered_landmarks.append(k)
                    landm_xyz[k] = v
                else:
                    raise Exception("Can't parse landmark %s: %s" % (k, v))
            if len(filtered_landmarks) > 0:
                import warnings
                warnings.warn("WARNING: the following landmarks are positioned at (0.0, 0.0, 0.0) and were ignored: %s" % ", ".join(filtered_landmarks))
            # We preserve these and calculate everything seperately so that we can recompute_landmarks if v changes
            self._raw_landmarks = {
                'landm': landm,
                'landm_xyz': landm_xyz
            }
            self.recompute_landmarks()

    def recompute_landmarks(self):
        import numpy as np
        landm = self._raw_landmarks['landm']
        landm_xyz = self._raw_landmarks['landm_xyz']
        self._landm = dict(landm.items() + self.compute_landmark_indices(landm_xyz).items())
        if len(self._landm) == 0:
            self._landm = None
            self._landm_regressors = None
        # compute default regressors from landmarks
        elif self.f is not None and len(self.f):
            landmark_points = np.vstack((self.v[landm.values()].reshape(-1, 3), np.array(landm_xyz.values()).reshape(-1, 3)))
            face_indices, closest_points = self.closest_faces_and_points(landmark_points)
            vertex_indices, coefficients = self.barycentric_coordinates_for_points(closest_points, face_indices)
            self._landm_regressors = {name: (vertex_indices[i], coefficients[i]) for i, name in enumerate(landm.keys() + landm_xyz.keys())}
        else:
            self._landm_regressors = {k: (v, np.array([1.0])) for k, v in self.landm.items()}

    @property
    def landm_names(self):
        if self.landm_regressors is not None:
            return self.landm_regressors.keys()
        elif self.landm is not None:
            return self.landm.keys()
        else:
            return []

    def compute_landmark_indices(self, landm_xyz):
        import numpy as np
        if len(landm_xyz) > 0:
            closest_vertices, _ = self.closest_vertices(np.array(landm_xyz.values()))
            return dict(zip(landm_xyz.keys(), closest_vertices))
        else:
            return {}

    def set_landmarks_from_xyz(self, landm_raw_xyz):
        raise DeprecationWarning("Mesh.set_landmarks_from_xyz is deprecated in favor of just setting mesh.landm_xyz directly")

    def set_landmark_indices_from_any(self, landmarks):
        raise DeprecationWarning("Mesh.set_landmark_indices_from_any is deprecated in favor of just setting mesh.landm directly")

    def set_landmarks_from_raw(self, landmarks):
        raise DeprecationWarning("Mesh.set_landmarks_from_raw is deprecated in favor of just setting mesh.landm directly")

    def set_landmarks_from_regressors(self, regressors):
        raise DeprecationWarning("Mesh.set_landmarks_from_regressors is deprecated in favor of just setting mesh.landm_regressors directly")




class MeshMixin(object):

    msg_lace_search_missing = "{} is implemented in `lace-search` which is not installed and not yet publicly avaliable"

    def compute_aabb_tree(self):
        try:
            from lace_search.aabb_tree import AabbTree
        except ImportError:
            raise NotImplementedError(self.msg_lace_search_missing.format('compute_aabb_tree'))
        return AabbTree(self.v, self.f)

    def compute_aabb_normals_tree(self):
        try:
            from lace_search.aabb_normals_tree import AabbNormalsTree
        except ImportError:
            raise NotImplementedError(self.msg_lace_search_missing.format('compute_aabb_normals_tree'))
        return AabbNormalsTree(self)

    def compute_closest_point_tree(self):
        try:
            from lace_search.closest_point_tree import ClosestPointTree
        except ImportError:
            raise NotImplementedError(self.msg_lace_search_missing.format('compute_closest_point_tree'))
        return ClosestPointTree(self)

    def closest_vertices(self, vertices):
        return self.compute_closest_point_tree().nearest(vertices)

    def closest_points(self, vertices):
        return self.closest_faces_and_points(vertices)[1]

    def closest_faces_and_points(self, vertices):
        return self.compute_aabb_tree().nearest(vertices)

    def vertices_within(self, vertex_or_vertices, radius):
        tree = self.compute_closest_point_tree()
        return tree.vertices_within(vertex_or_vertices, radius)



class MeshMixin(object):
    def show(self, mv=None, meshes=None, lines=None, with_measurements=False, titlebar='Mesh Viewer'):
        import numpy as np
        from lace.meshviewer import MeshViewer
        from blmath.numerics.matlab import row
        if meshes is None:
            meshes = []
        if lines is None:
            lines = []
        if mv is None:
            mv = MeshViewer(keepalive=True, titlebar=titlebar)
        else:
            mv.titlebar = titlebar
        if self.landm is not None:
            from lace.sphere import Sphere
            sphere = Sphere()
            scalefactor = 1e-2 * np.max(np.max(self.v) - np.min(self.v)) / np.max(np.max(sphere.v) - np.min(sphere.v))
            sphere.scale(scalefactor)
            spheres = [self.__class__(vc='SteelBlue', f=sphere.f, v=sphere.v + row(np.array(self.landm_xyz[k]))) for k in self.landm.keys()]
            mv.set_dynamic_meshes([self] + spheres + meshes, blocking=True)
        else:
            mv.set_dynamic_meshes([self] + meshes, blocking=True)
        if with_measurements:
            lines.extend([line for m in self.measurements for line in m.lines])
        mv.set_dynamic_lines(lines)
        return mv



# pylint: disable=attribute-defined-outside-init, access-member-before-definition, len-as-condition
from cached_property import cached_property

def quads_to_tris(quads):
    '''
    Convert quad faces to triangular faces.

    quads: An nx4 array.

    Return a 2nx3 array.

    '''
    import numpy as np
    tris = np.empty((2 * len(quads), 3))
    tris[0::2, :] = quads[:, [0, 1, 2]]
    tris[1::2, :] = quads[:, [0, 2, 3]]
    return tris


def vertices_in_common(face_1, face_2):
    "returns the two vertices shared by two faces"
    # So... on a 10000 iteration timing run, np.intersect1d takes 0.2s, the very
    # complicated code we pulled from core takes 0.05s, and this runs in 0.007s.
    # Just goes to show that sometimes simplest is best... The easy timing script
    # for things like this is:
    # import timeit; print timeit.timeit('vertices_in_common([0, 1, 2], [0, 1, 3])', setup='from lace.topology import vertices_in_common', number=10000)
    return [x for x in face_1 if x in face_2]


class MeshMixin(object):
    def faces_by_vertex(self, as_sparse_matrix=False):
        import numpy as np
        import scipy.sparse as sp
        if not as_sparse_matrix:
            faces_by_vertex = [[] for i in range(len(self.v))]
            for i, face in enumerate(self.f):
                faces_by_vertex[face[0]].append(i)
                faces_by_vertex[face[1]].append(i)
                faces_by_vertex[face[2]].append(i)
        else:
            row = self.f.flatten()
            col = np.array([range(self.f.shape[0])]*3).T.flatten()
            data = np.ones(len(col))
            faces_by_vertex = sp.csr_matrix((data, (row, col)), shape=(self.v.shape[0], self.f.shape[0]))
        return faces_by_vertex

    def all_faces_with_verts(self, v_indices, as_boolean=False):
        '''
        returns all of the faces that contain at least one of the vertices in v_indices
        '''
        import numpy as np
        included_vertices = np.zeros(self.v.shape[0], dtype=bool)
        included_vertices[np.array(v_indices, dtype=np.uint32)] = True
        faces_with_verts = included_vertices[self.f].all(axis=1)
        if as_boolean:
            return faces_with_verts
        return np.nonzero(faces_with_verts)[0]

    def transfer_segm(self, mesh, exclude_empty_parts=True):
        import numpy as np
        self.segm = {}
        if mesh.segm is not None:
            face_centers = np.array([self.v[face, :].mean(axis=0) for face in self.f])
            closest_faces, _ = mesh.closest_faces_and_points(face_centers)
            mesh_parts_by_face = mesh.parts_by_face()
            parts_by_face = [mesh_parts_by_face[face] for face in closest_faces.flatten()]
            self.segm = dict([(part, []) for part in mesh.segm.keys()])
            for face, part in enumerate(parts_by_face):
                if part:
                    self.segm[part].append(face)
            for part in self.segm.keys():
                self.segm[part].sort()
                if exclude_empty_parts and not self.segm[part]:
                    del self.segm[part]

    def vertex_indices_in_segments(self, segments, ret_face_indices=False):
        '''
        Given a list of segment names, return an array of vertex indices for
        all the vertices in those faces.

        Args:
            segments: a list of segment names,
            ret_face_indices: if it is `True`, returns face indices
        '''
        import numpy as np
        import warnings

        face_indices = np.array([])
        vertex_indices = np.array([])
        if self.segm is not None:
            try:
                segments = [self.segm[name] for name in segments]
            except KeyError as e:
                raise ValueError('Unknown segments {}. Consier using Mesh.clean_segments on segments'.format(e.args[0]))
            face_indices = np.unique(np.concatenate(segments))
            vertex_indices = np.unique(np.ravel(self.f[face_indices]))
        else:
            warnings.warn('self.segm is None, will return empty array')

        if ret_face_indices:
            return vertex_indices, face_indices
        else:
            return vertex_indices

    def clean_segments(self, segments):
        """Return a list of segments that are in self.segm
        """
        return [name for name in segments if name in self.segm]

    def keep_segments(self, segments_to_keep, preserve_segmentation=True):
        '''
        Keep the faces and vertices for given segments, discarding all others.
        When preserve_segmentation is false self.segm is discarded for speed.
        '''
        v_ind, f_ind = self.vertex_indices_in_segments(segments_to_keep, ret_face_indices=True)
        self.segm = {name: self.segm[name] for name in segments_to_keep}

        if not preserve_segmentation:
            self.segm = None
            self.f = self.f[f_ind]
            if self.ft is not None:
                self.ft = self.ft[f_ind]

        self.keep_vertices(v_ind)

    def remove_segments(self, segments_to_remove):
        ''' Remove the faces and vertices for given segments, keeping all others.

        Args:
            segments_to_remove: a list of segnments whose vertices will be removed
        '''
        v_ind = self.vertex_indices_in_segments(segments_to_remove)
        self.segm = {name: faces for name, faces in self.segm.iteritems() if name not in segments_to_remove}
        self.remove_vertices(v_ind)

    @property
    def verts_by_segm(self):
        return dict((segment, sorted(set(self.f[indices].flatten()))) for segment, indices in self.segm.items())

    def parts_by_face(self):
        segments_by_face = ['']*len(self.f)
        for part in self.segm.keys():
            for face in self.segm[part]:
                segments_by_face[face] = part
        return segments_by_face

    def verts_in_common(self, segments):
        """
        returns array of all vertex indices common to each segment in segments"""
        verts_by_segm = self.verts_by_segm
        return sorted(reduce(lambda s0, s1: s0.intersection(s1),
                             [set(verts_by_segm[segm]) for segm in segments]))
        ## indices of vertices in the faces of the first segment
        #indices = self.verts_by_segm[segments[0]]
        #for segment in segments[1:]:
        #    indices = sorted([index for index in self.verts_by_segm[segment] if index in indices]) # Intersect current segment with current indices
        #return sorted(set(indices))

    @property
    def joint_names(self):
        return self.joint_regressors.keys()

    @property
    def joint_xyz(self):
        import numpy as np
        joint_locations = {}
        for name in self.joint_names:
            joint_locations[name] = self.joint_regressors[name]['offset'] + np.sum(self.v[self.joint_regressors[name]['v_indices']].T*self.joint_regressors[name]['coeff'], axis=1)
        return joint_locations

    # creates joint_regressors from a list of joint names and a per joint list of vertex indices (e.g. a ring of vertices)
    # For the regression coefficients, all vertices for a given joint are given equal weight
    def set_joints(self, joint_names, vertex_indices):
        import numpy as np
        self.joint_regressors = {}
        for name, indices in zip(joint_names, vertex_indices):
            self.joint_regressors[name] = {'v_indices': indices, 'coeff': [1.0/len(indices)]*len(indices), 'offset': np.array([0., 0., 0.])}

    def uniquified_mesh(self):
        """This function returns a copy of the mesh in which vertices are copied such that
        each vertex appears in only one face, and hence has only one texture"""
        import numpy as np
        from lace.mesh import Mesh
        new_mesh = Mesh(v=self.v[self.f.flatten()], f=np.array(range(len(self.f.flatten()))).reshape(-1, 3))

        if self.vn is None:
            self.reset_normals()
        new_mesh.vn = self.vn[self.f.flatten()]

        if self.vt is not None:
            new_mesh.vt = self.vt[self.ft.flatten()]
            new_mesh.ft = new_mesh.f.copy()
        return new_mesh

    def downsampled_mesh(self, step):
        """Returns a downsampled copy of this mesh.

        Args:
            step: the step size for the sampling

        Returns:
            a new, downsampled Mesh object.

        Raises:
            ValueError if this Mesh has faces.
        """
        from lace.mesh import Mesh

        if self.f is not None:
            raise ValueError(
                'Function `downsampled_mesh` does not support faces.')

        low = Mesh()
        if self.v is not None:
            low.v = self.v[::step]
        if self.vc is not None:
            low.vc = self.vc[::step]
        return low

    def keep_vertices(self, indices_to_keep, ret_kept_faces=False):
        '''
        Keep the given vertices and discard the others, and any faces to which
        they may belong.


        If `ret_kept_faces` is `True`, return the original indices of the kept
        faces. Otherwise return `self` for chaining.

        '''
        import numpy as np

        if self.v is None:
            return

        indices_to_keep = np.array(indices_to_keep, dtype=np.uint32)

        initial_num_verts = self.v.shape[0]
        if self.f is not None:
            initial_num_faces = self.f.shape[0]
            f_indices_to_keep = self.all_faces_with_verts(indices_to_keep, as_boolean=True)

        # Why do we test this? Don't know. But we do need to test it before we
        # mutate self.v.
        vn_should_update = self.vn is not None and self.vn.shape[0] == initial_num_verts
        vc_should_update = self.vc is not None and self.vc.shape[0] == initial_num_verts

        self.v = self.v[indices_to_keep]

        if vn_should_update:
            self.vn = self.vn[indices_to_keep]
        if vc_should_update:
            self.vc = self.vc[indices_to_keep]

        if self.f is not None:
            v_old_to_new = np.zeros(initial_num_verts, dtype=int)
            f_old_to_new = np.zeros(initial_num_faces, dtype=int)

            v_old_to_new[indices_to_keep] = np.arange(len(indices_to_keep), dtype=int)
            self.f = v_old_to_new[self.f[f_indices_to_keep]]
            f_old_to_new[f_indices_to_keep] = np.arange(self.f.shape[0], dtype=int)

        else:
            # Make the code below work, in case there is somehow degenerate
            # segm even though there are no faces.
            f_indices_to_keep = []

        if self.segm is not None:
            new_segm = {}
            for segm_name, segm_faces in self.segm.items():
                faces = np.array(segm_faces, dtype=int)
                valid_faces = faces[f_indices_to_keep[faces]]
                if len(valid_faces):
                    new_segm[segm_name] = f_old_to_new[valid_faces]
            self.segm = new_segm if new_segm else None

        if hasattr(self, '_raw_landmarks') and self._raw_landmarks is not None:
            self.recompute_landmarks()

        return np.nonzero(f_indices_to_keep)[0] if ret_kept_faces else self

    def remove_vertices(self, v_list):
        import numpy as np
        return self.keep_vertices(np.setdiff1d(np.arange(self.v.shape[0]), v_list))

    def point_cloud(self):
        return self.__class__(v=self.v, f=[], vc=self.vc) if self.vc is not None else self.__class__(v=self.v, f=[])

    def remove_faces(self, face_indices_to_remove):
        import numpy as np
        self.f = np.delete(self.f, face_indices_to_remove, 0)
        return self

    def flip_faces(self):
        self.f = self.f.copy()
        for i in range(len(self.f)):
            self.f[i] = self.f[i][::-1]
        if self.ft is not None:
            for i in range(len(self.f)):
                self.ft[i] = self.ft[i][::-1]
        return self

    def subdivide_triangles(self):
        import numpy as np
        new_faces = []
        new_vertices = self.v.copy()
        for face in self.f:
            face_vertices = np.array([self.v[face[0], :], self.v[face[1], :], self.v[face[2], :]])
            new_vertex = np.mean(face_vertices, axis=0)
            new_vertices = np.vstack([new_vertices, new_vertex])
            new_vertex_index = len(new_vertices) - 1
            if len(new_faces):
                new_faces = np.vstack([new_faces, [face[0], face[1], new_vertex_index], [face[1], face[2], new_vertex_index], [face[2], face[0], new_vertex_index]])
            else:
                new_faces = np.array([[face[0], face[1], new_vertex_index], [face[1], face[2], new_vertex_index], [face[2], face[0], new_vertex_index]])
        self.v = new_vertices
        self.f = new_faces

        if self.vt is not None:
            new_ft = []
            new_texture_coordinates = self.vt.copy()
            for face_texture in self.ft:
                face_texture_coordinates = np.array([self.vt[face_texture[0], :], self.vt[face_texture[1], :], self.vt[face_texture[2], :]])
                new_texture_coordinate = np.mean(face_texture_coordinates, axis=0)
                new_texture_coordinates = np.vstack([new_texture_coordinates, new_texture_coordinate])
                new_texture_index = len(new_texture_coordinates) - 1
                if len(new_ft):
                    new_ft = np.vstack([new_ft, [face_texture[0], face_texture[1], new_texture_index], [face_texture[1], face_texture[2], new_texture_index], [face_texture[2], face_texture[0], new_texture_index]])
                else:
                    new_ft = np.array([[face_texture[0], face_texture[1], new_texture_index], [face_texture[1], face_texture[2], new_texture_index], [face_texture[2], face_texture[0], new_texture_index]])
            self.vt = new_texture_coordinates
            self.ft = new_ft
        return self

    def concatenate_mesh(self, mesh):
        import numpy as np
        if len(self.v) == 0:
            self.f = mesh.f.copy()
            self.v = mesh.v.copy()
            self.vc = mesh.vc.copy() if mesh.vc is not None else None
        elif len(mesh.v):
            self.f = np.concatenate([self.f, mesh.f.copy() + len(self.v)])
            self.v = np.concatenate([self.v, mesh.v])
            self.vc = np.concatenate([self.vc, mesh.vc]) if (mesh.vc is not None and self.vc is not None) else None
        return self

    # new_ordering specifies the new index of each vertex. If new_ordering[i] = j,
    # vertex i should now be the j^th vertex. As such, each entry in new_ordering should be unique.
    def reorder_vertices(self, new_ordering, new_normal_ordering=None):
        import numpy as np
        if new_normal_ordering is None:
            new_normal_ordering = new_ordering
        inverse_ordering = np.zeros(len(new_ordering), dtype=int)
        for i, j in enumerate(new_ordering):
            inverse_ordering[j] = i
        inverse_normal_ordering = np.zeros(len(new_normal_ordering), dtype=int)
        for i, j in enumerate(new_normal_ordering):
            inverse_normal_ordering[j] = i
        self.v = self.v[inverse_ordering]
        if self.vn is not None:
            self.vn = self.vn[inverse_normal_ordering]
        for i in range(len(self.f)):
            self.f[i] = np.array([new_ordering[vertex_index] for vertex_index in self.f[i]])
            if self.fn is not None:
                self.fn[i] = np.array([new_normal_ordering[normal_index] for normal_index in self.fn[i]])

    @classmethod
    def create_from_mesh_and_lines(cls, mesh, lines):
        '''
        Return a copy of mesh with line vertices and edges added.

        mesh: A Mesh
        lines: A list of Polyline or Lines objects.

        '''
        mesh_with_lines = mesh.copy()
        mesh_with_lines.add_lines(lines)
        return mesh_with_lines

    def add_lines(self, lines):
        '''
        Add line vertices and edges to the mesh.

        lines: A list of Polyline or Lines objects.

        '''
        import numpy as np
        if not lines:
            return

        v_lines = np.vstack([l.v for l in lines])
        v_index_offset = np.cumsum([0] + [len(l.v) for l in lines])
        e_lines = np.vstack([l.e + v_index_offset[i] for i, l in enumerate(lines)])
        num_body_verts = self.v.shape[0]
        self.v = np.vstack([self.v, v_lines])
        self.e = e_lines + num_body_verts

    @property
    def vert_connectivity(self):
        """Returns a sparse matrix (of size #verts x #verts) where each nonzero
        element indicates a neighborhood relation. For example, if there is a
        nonzero element in position (15, 12), that means vertex 15 is connected
        by an edge to vertex 12."""
        import numpy as np
        import scipy.sparse as sp
        from blmath.numerics.matlab import row
        vpv = sp.csc_matrix((len(self.v), len(self.v)))
        # for each column in the faces...
        for i in range(3):
            IS = self.f[:, i]
            JS = self.f[:, (i+1)%3]
            data = np.ones(len(IS))
            ij = np.vstack((row(IS.flatten()), row(JS.flatten())))
            mtx = sp.csc_matrix((data, ij), shape=vpv.shape)
            vpv = vpv + mtx + mtx.T
        return vpv

    @property
    def vert_opposites_per_edge(self):
        """Returns a dictionary from vertidx-pairs to opposites.
        For example, a key consist of [4, 5)] meaning the edge between
        vertices 4 and 5, and a value might be [10, 11] which are the indices
        of the vertices opposing this edge."""
        result = {}
        for f in self.f:
            for i in range(3):
                key = [f[i], f[(i+1)%3]]
                key.sort()
                key = tuple(key)
                val = f[(i+2)%3]

                if key in result:
                    result[key].append(val)
                else:
                    result[key] = [val]
        return result

    @cached_property
    def faces_per_edge(self):
        """Returns an Ex2 array of adjacencies between faces, where
        each element in the array is a face index. Each edge is included
        only once. Edges that are not shared by 2 faces are not included."""
        import numpy as np
        import scipy.sparse as sp
        from blmath.numerics.matlab import col
        IS = np.repeat(np.arange(len(self.f)), 3)
        JS = self.f.ravel()
        data = np.ones(IS.size)
        f2v = sp.csc_matrix((data, (IS, JS)), shape=(len(self.f), np.max(self.f.ravel())+1))
        f2f = f2v.dot(f2v.T)
        f2f = f2f.tocoo()
        f2f = np.hstack((col(f2f.row), col(f2f.col), col(f2f.data)))
        which = (f2f[:, 0] < f2f[:, 1]) & (f2f[:, 2] >= 2)
        return np.asarray(f2f[which, :2], np.uint32)

    @cached_property
    def vertices_per_edge(self):
        """Returns an Ex2 array of adjacencies between vertices, where
        each element in the array is a vertex index. Each edge is included
        only once. Edges that are not shared by 2 faces are not included."""
        import numpy as np
        return np.asarray([vertices_in_common(e[0], e[1]) for e in self.f[self.faces_per_edge]])

    def get_vertices_to_edges_matrix(self, want_xyz=True):
        """Returns a matrix M, which if multiplied by vertices,
        gives back edges (so "e = M.dot(v)"). Note that this generates
        one edge per edge, *not* two edges per triangle.

        Args:
            want_xyz: if true, takes and returns xyz coordinates, otherwise
                takes and returns x *or* y *or* z coordinates
        """
        import numpy as np
        import scipy.sparse as sp

        vpe = np.asarray(self.vertices_per_edge, dtype=np.int32)
        IS = np.repeat(np.arange(len(vpe)), 2)
        JS = vpe.flatten()
        data = np.ones_like(vpe)
        data[:, 1] = -1
        data = data.flatten()

        if want_xyz:
            IS = np.concatenate((IS*3, IS*3+1, IS*3+2))
            JS = np.concatenate((JS*3, JS*3+1, JS*3+2))
            data = np.concatenate((data, data, data))

        ij = np.vstack((IS.flatten(), JS.flatten()))
        return sp.csc_matrix((data, ij))

    @cached_property
    def vertices_to_edges_matrix(self):
        return self.get_vertices_to_edges_matrix(want_xyz=True)

    @cached_property
    def vertices_to_edges_matrix_single_axis(self):
        return self.get_vertices_to_edges_matrix(want_xyz=False)

    def remove_redundant_verts(self, eps=1e-10):
        """Given verts and faces, this remove colocated vertices"""
        import numpy as np
        from scipy.spatial import cKDTree # FIXME pylint: disable=no-name-in-module
        fshape = self.f.shape
        tree = cKDTree(self.v)
        close_pairs = list(tree.query_pairs(eps))
        if close_pairs:
            close_pairs = np.sort(close_pairs, axis=1)
            # update faces to not refer to redundant vertices
            equivalent_verts = np.arange(self.v.shape[0])
            for v1, v2 in close_pairs:
                if equivalent_verts[v2] > v1:
                    equivalent_verts[v2] = v1
            self.f = equivalent_verts[self.f.flatten()].reshape((-1, 3))
            # get rid of unused verts, and update faces accordingly
            vertidxs_left = np.unique(self.f)
            repl = np.arange(np.max(self.f)+1)
            repl[vertidxs_left] = np.arange(len(vertidxs_left))
            self.v = self.v[vertidxs_left]
            self.f = repl[self.f].reshape((-1, fshape[1]))



import math
import numpy as np
from lace.mesh import Mesh


def create_rectangular_prism(origin, size):
    '''
    Return a Mesh which is an axis-aligned rectangular prism. One vertex is
    `origin`; the diametrically opposite vertex is `origin + size`.

    size: 3x1 array.

    '''
    from lace.topology import quads_to_tris

    lower_base_plane = np.array([
        # Lower base plane
        origin,
        origin + np.array([size[0], 0, 0]),
        origin + np.array([size[0], 0, size[2]]),
        origin + np.array([0, 0, size[2]]),
    ])
    upper_base_plane = lower_base_plane + np.array([0, size[1], 0])

    vertices = np.vstack([lower_base_plane, upper_base_plane])

    faces = quads_to_tris(np.array([
        [0, 1, 2, 3],  # lower base (-y)
        [7, 6, 5, 4],  # upper base (+y)
        [4, 5, 1, 0],  # +z face
        [5, 6, 2, 1],  # +x face
        [6, 7, 3, 2],  # -z face
        [3, 7, 4, 0],  # -x face
    ]))

    return Mesh(v=vertices, f=faces)


def create_cube(origin, size):
    '''
    Return a mesh with an axis-aligned cube. One vertex is `origin`; the
    diametrically opposite vertex is `size` units along +x, +y, and +z.

    size: int or float.

    '''
    return create_rectangular_prism(origin, np.repeat(size, 3))


def create_triangular_prism(p1, p2, p3, height):
    '''
    Return a Mesh which is a triangular prism whose base is the triangle
    p1, p2, p3. If the vertices are oriented in a counterclockwise
    direction, the prism extends from behind them.

    '''
    from blmath.geometry import Plane

    base_plane = Plane.from_points(p1, p2, p3)
    lower_base_to_upper_base = height * -base_plane.normal # pylint: disable=invalid-unary-operand-type
    vertices = np.vstack(([p1, p2, p3], [p1, p2, p3] + lower_base_to_upper_base))

    faces = np.array([
        [0, 1, 2],  # base
        [0, 3, 4], [0, 4, 1],  # side 0, 3, 4, 1
        [1, 4, 5], [1, 5, 2],  # side 1, 4, 5, 2
        [2, 5, 3], [2, 3, 0],  # side 2, 5, 3, 0
        [5, 4, 3],  # base
    ])

    return Mesh(v=vertices, f=faces)


def create_horizontal_plane():
    '''
    Creates a horizontal plane.
    '''
    v = np.array([
        [1., 0., 0.],
        [-1., 0., 0.],
        [0., 0., 1.],
        [0., 0., -1.]
    ])
    f = [[0, 1, 2], [3, 1, 0]]
    return Mesh(v=v, f=f)


if __name__ == '__main__':
    points = np.array([
        [1, 0, 0],
        [0, math.sqrt(1.25), 0],
        [-1, 0, 0],
    ])
    prism = create_triangular_prism(*points, height=4)
    prism.show()

    cube = create_cube(np.array([1., 0., 0.]), 4.)
    cube.show()



'''
Reads and writes a Meshlab picked points file. Files contain named
points in 3D space. A set of picked points can be saved alongside
the mesh and loaded in Meshlab (on the Edit menu, click PickPoints),
or via Mesh(filename=mesh_filename, landmarks=pp_filename).

'''

from __future__ import absolute_import
__all__ = ['load', 'dump', 'loads', 'dumps', 'EXTENSION']

EXTENSION = '.pp'


def dump(obj, f, *args, **kwargs):
    from baiji.serialization.util.openlib import ensure_file_open_and_call
    return ensure_file_open_and_call(f, _dump, 'w', obj, *args, **kwargs)


def load(f, *args, **kwargs):
    from baiji.serialization.util.openlib import ensure_file_open_and_call
    return ensure_file_open_and_call(f, _load, 'r', *args, **kwargs)


def loads(s, *args, **kwargs): # pylint: disable=unused-argument
    # TODO This seems tricky with xml.etree, probably easiest to
    # write the file and then read it again?
    raise NotImplementedError()


def dumps(obj, mesh_filename=None, *args, **kwargs): # pylint: disable=unused-argument
    '''
    obj: A dictionary mapping names to a 3-dimension array.
    mesh_filename: If provided, this value is included in the <DataFileName>
      attribute, which Meshlab doesn't seem to use.

    TODO Maybe reconstruct this using xml.etree

    '''
    point_template = '<point x="%f" y="%f" z="%f" name="%s"/>\n'
    file_template = """
    <!DOCTYPE PickedPoints>
    <PickedPoints>
     <DocumentData>
      <DateTime time="16:00:00" date="2014-12-31"/>
      <User name="bodylabs"/>
      <DataFileName name="%s"/>
     </DocumentData>
     %s
    </PickedPoints>
    """
    from blmath.numerics import isnumericarray

    if not isinstance(obj, dict) or not all([isnumericarray(point) for point in obj.itervalues()]):
        raise ValueError('obj should be a dict of points')

    points = '\n'.join([point_template % (tuple(xyz) + (name,)) for name, xyz in obj.iteritems()])

    return file_template % (mesh_filename, points)


def _dump(f, obj, *args, **kwargs):
    xml_string = dumps(obj, *args, **kwargs)
    f.write(xml_string)


def _load(f, *args, **kwargs): # pylint: disable=unused-argument
    from xml.etree import ElementTree
    tree = ElementTree.parse(f)

    points = {}

    for e in tree.iter('point'):
        try:
            point = [float(e.attrib['x']), float(e.attrib['y']), float(e.attrib['z'])]
        except ValueError: # may happen if landmarks are just spaces
            point = [0, 0, 0]
        points[e.attrib['name']] = point

    return points



def load(f, existing_mesh=None):
    from baiji.serialization.util.openlib import ensure_file_open_and_call
    return ensure_file_open_and_call(f, _load, mode='rb', existing_mesh=existing_mesh)

def _load(f, existing_mesh=None):
    import numpy as np
    from lace.mesh import Mesh
    v = np.loadtxt(f.name)
    if existing_mesh is None:
        return Mesh(v=v)
    else:
        existing_mesh.v = v
        return existing_mesh



import unittest
import numpy as np
from lace.mesh import Mesh
from lace import color

class TestMeshColors(unittest.TestCase):
    def setUp(self):
        self.colormap = 'jet'
        self.jet_0 = np.array([0.0, 0.0, 0.5])
        self.jet_1 = np.array([0.5, 0.0, 0.0])

    def test_colors_like_with_full_array(self):
        c = np.random.rand(10, 3)
        v = np.ones((10, 3))
        vc = color.colors_like(c, v)
        np.testing.assert_array_equal(c, vc)
        self.assertEqual(vc.dtype, np.float64)

    def test_colors_like_with_full_array_transposed(self):
        c = np.random.rand(10, 3)
        v = np.ones((10, 3))
        vc = color.colors_like(c.T, v)
        np.testing.assert_array_equal(c, vc)
        self.assertEqual(vc.dtype, np.float64)

    def test_colors_like_with_single_row(self):
        c = np.array([[1.0, 0.0, 1.0, 0.0]])
        v = np.ones((4, 3))
        vc = color.colors_like(c, v, colormap=self.colormap)
        expected_c = np.array([self.jet_1, self.jet_0, self.jet_1, self.jet_0])
        np.testing.assert_array_equal(expected_c, vc)
        self.assertEqual(vc.dtype, np.float64)

    def test_colors_like_with_single_col(self):
        c = np.array([[1.0], [0.0], [1.0], [0.0]])
        v = np.ones((4, 3))
        vc = color.colors_like(c, v, colormap=self.colormap)
        expected_c = np.array([self.jet_1, self.jet_0, self.jet_1, self.jet_0])
        np.testing.assert_array_equal(expected_c, vc)
        self.assertEqual(vc.dtype, np.float64)

    def test_colors_like_with_array(self):
        c = np.array([1.0, 0.0, 1.0, 0.0])
        v = np.ones((4, 3))
        vc = color.colors_like(c, v, colormap=self.colormap)
        expected_c = np.array([self.jet_1, self.jet_0, self.jet_1, self.jet_0])
        np.testing.assert_array_equal(expected_c, vc)
        self.assertEqual(vc.dtype, np.float64)

    def test_colors_like_with_list(self):
        c = [1.0, 0.0, 1.0, 0.0]
        v = np.ones((4, 3))
        vc = color.colors_like(c, v, colormap=self.colormap)
        expected_c = np.array([self.jet_1, self.jet_0, self.jet_1, self.jet_0])
        np.testing.assert_array_equal(expected_c, vc)
        self.assertEqual(vc.dtype, np.float64)

    def test_colors_like_with_name(self):
        c = 'red'
        v = np.ones((4, 3))
        vc = color.colors_like(c, v, colormap=self.colormap)
        expected_c = np.array([1.0, 0.0, 0.0]) * v
        np.testing.assert_array_equal(expected_c, vc)
        self.assertEqual(vc.dtype, np.float64)

    def test_colors_like_rgb_triple_as_array(self):
        c = [1.0, 0.0, 1.0]
        v = np.ones((4, 3))
        vc = color.colors_like(c, v, colormap=self.colormap)
        expected_c = np.array([1.0, 0.0, 1.0]) * v
        np.testing.assert_array_equal(expected_c, vc)
        self.assertEqual(vc.dtype, np.float64)

    def test_colors_like_rgb_triple_as_col(self):
        c = [[1.0], [0.0], [1.0]]
        v = np.ones((4, 3))
        vc = color.colors_like(c, v, colormap=self.colormap)
        expected_c = np.array([1.0, 0.0, 1.0]) * v
        np.testing.assert_array_equal(expected_c, vc)
        self.assertEqual(vc.dtype, np.float64)

    def test_colors_like_rgb_triple_as_row(self):
        c = [[1.0, 0.0, 1.0]]
        v = np.ones((4, 3))
        vc = color.colors_like(c, v, colormap=self.colormap)
        expected_c = np.array([1.0, 0.0, 1.0]) * v
        np.testing.assert_array_equal(expected_c, vc)
        self.assertEqual(vc.dtype, np.float64)

    def test_colors_like_none_or_empty(self):
        for c in [None, [], (), np.array([])]:
            v = np.ones((4, 3))
            vc = color.colors_like(c, v, colormap=self.colormap)
            self.assertIsNone(vc)

    def test_setting_vc_in_mesh_constructor(self):
        m = Mesh(v=np.ones((4, 3)), vc='red')
        expected_c = np.array([1.0, 0.0, 0.0]) * m.v
        np.testing.assert_array_equal(expected_c, m.vc)
        self.assertEqual(m.vc.dtype, np.float64)

    def test_scale_vertex_colors(self):
        m = Mesh(v=np.ones((4, 3)), vc='red')
        m.scale_vertex_colors(np.array([2.0, 1.0, 0.5, 0.0]))
        expected_c = np.array([[1.0, 0.0, 0.0], [0.5, 0.0, 0.0], [0.25, 0.0, 0.0], [0.0, 0.0, 0.0]])
        np.testing.assert_array_equal(expected_c, m.vc)
        self.assertEqual(m.vc.dtype, np.float64)



# pylint: disable=attribute-defined-outside-init
class MeshMixin(object):

    @property
    def texture_filepath(self):
        if getattr(self, '_texture_filepath', None) is None:
            return None
        if self._texture_filepath.startswith('/is/ps/'):
            # This is a special case to support some old meshes, scape templates in particular,
            # that have old MPI paths encoded in them. Note that we do this here rather than in
            # the setter, since some of them we're loading from pickle files and unpickling doesn't
            # call the setter.
            return "s3://bodylabs-korper-assets" + self._texture_filepath
        return self._texture_filepath

    @texture_filepath.setter
    def texture_filepath(self, val):
        self._texture_image = None # To ensure reloading
        self._texture_filepath = val

    @property
    def texture_image(self):
        if getattr(self, '_texture_image', None) is None:
            self.reload_texture_image()
        return self._texture_image

    def set_texture_image(self, path_to_texture):
        self.texture_filepath = path_to_texture

    def texture_coordinates_by_vertex(self):
        texture_coordinates_by_vertex = [[] for i in range(len(self.v))]
        for i, face in enumerate(self.f):
            for j in [0, 1, 2]:
                texture_coordinates_by_vertex[face[j]].append(self.vt[self.ft[i][j]])
        return texture_coordinates_by_vertex


    def reload_texture_image(self):
        import cv2
        import numpy as np
        from baiji import s3
        if not self.texture_filepath:
            self._texture_image = None
        else:
            # image is loaded as image_height-by-image_width-by-3 array in BGR color order.
            with s3.open(self.texture_filepath) as f:
                self._texture_image = cv2.imread(f.name)
            texture_sizes = [32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384]
            if self._texture_image is not None:
                h, w = self._texture_image.shape[:2]
                if h != w or h not in texture_sizes or w not in texture_sizes:
                    closest_texture_size_idx = (np.abs(np.array(texture_sizes) - max(self._texture_image.shape))).argmin()
                    sz = texture_sizes[closest_texture_size_idx]
                    self._texture_image = cv2.resize(self._texture_image, (sz, sz))

    def load_texture(self, texture_version):
        '''
        Expect a texture version number as an integer, load the texture version from /is/ps/shared/data/body/template/texture_coordinates/.
        Currently there are versions [0, 1, 2, 3] availiable.
        '''
        import numpy as np
        lowres_tex_template = 's3://bodylabs-korper-assets/is/ps/shared/data/body/template/texture_coordinates/textured_template_low_v%d.obj' % texture_version
        highres_tex_template = 's3://bodylabs-korper-assets/is/ps/shared/data/body/template/texture_coordinates/textured_template_high_v%d.obj' % texture_version
        from lace.mesh import Mesh
        from lace.cache import sc
        mesh_with_texture = Mesh(filename=sc(lowres_tex_template))
        if not np.all(mesh_with_texture.f.shape == self.f.shape):
            mesh_with_texture = Mesh(filename=sc(highres_tex_template))
        self.transfer_texture(mesh_with_texture)

    def transfer_texture(self, mesh_with_texture):
        import numpy as np
        if not np.all(mesh_with_texture.f.shape == self.f.shape):
            raise Exception('Mesh topology mismatch')
        self.vt = mesh_with_texture.vt.copy()
        self.ft = mesh_with_texture.ft.copy()
        if not np.all(mesh_with_texture.f == self.f):
            if np.all(mesh_with_texture.f == np.fliplr(self.f)):
                self.ft = np.fliplr(self.ft)
            else:
                # Same shape; let's see if it's face ordering; this could be a bit faster...
                face_mapping = {}
                for f, ii in zip(self.f, range(len(self.f))):
                    face_mapping[" ".join([str(x) for x in sorted(f)])] = ii
                self.ft = np.zeros(self.f.shape, dtype=np.uint32)
                for f, ft in zip(mesh_with_texture.f, mesh_with_texture.ft):
                    k = " ".join([str(x) for x in sorted(f)])
                    if k not in face_mapping:
                        raise Exception('Mesh topology mismatch')
                    # the vertex order can be arbitrary...
                    ids = []
                    for f_id in f:
                        ids.append(np.where(self.f[face_mapping[k]] == f_id)[0][0])
                    ids = np.array(ids)
                    self.ft[face_mapping[k]] = np.array(ft[ids])
        self.texture_filepath = mesh_with_texture.texture_filepath
        self._texture_image = None

    def texture_rgb(self, texture_coordinate):
        import numpy as np
        h, w = np.array(self.texture_image.shape[:2]) - 1
        return np.double(self.texture_image[int(h*(1.0 - texture_coordinate[1]))][int(w*(texture_coordinate[0]))])[::-1]

    def texture_rgb_vec(self, texture_coordinates):
        import numpy as np
        h, w = np.array(self.texture_image.shape[:2]) - 1
        n_ch = self.texture_image.shape[2]
        # XXX texture_coordinates can be lower than 0! clip needed!
        d1 = (h*(1.0 - np.clip(texture_coordinates[:, 1], 0, 1))).astype(np.int)
        d0 = (w*(np.clip(texture_coordinates[:, 0], 0, 1))).astype(np.int)
        flat_texture = self.texture_image.flatten()
        indices = np.hstack([((d1*(w+1)*n_ch) + (d0*n_ch) + (2-i)).reshape(-1, 1) for i in range(n_ch)])
        return flat_texture[indices]



import numpy as np
from blmath.numerics import as_numeric_array
from blmath.util.decorators import setter_property

__all__ = ['Lines']

class Lines(object):
    """3d List-of-lines

    Attributes:
        v: Vx3 array of vertices
        e: Ex2 array of edges
    """

    def __init__(self, v, e, vc=None, ec=None):
        self.v = v
        self.e = e
        self.vc = vc
        self.ec = ec

    @setter_property
    def v(self, val):  # pylint: disable=method-hidden
        self.__dict__['v'] = as_numeric_array(val, dtype=np.float64, shape=(-1, 3), allow_none=True, empty_as_none=True)

    @setter_property
    def e(self, val):  # pylint: disable=method-hidden
        self.__dict__['e'] = as_numeric_array(val, dtype=np.uint64, shape=(-1, 2), allow_none=True, empty_as_none=True)

    @setter_property
    def vc(self, val):  # pylint: disable=method-hidden
        from lace import color
        val = color.colors_like(val, self.v)
        self.__dict__['vc'] = val

    @setter_property
    def ec(self, val):  # pylint: disable=method-hidden
        from lace import color
        val = color.colors_like(val, self.e)
        self.__dict__['ec'] = val

    def write_obj(self, filename):
        with open(filename, 'w') as fi:
            for r in self.v:
                fi.write('v %f %f %f\n' % (r[0], r[1], r[2]))
            for e in self.e:
                fi.write('l %d %d\n' % (e[0]+1, e[1]+1))

    def show(self, mv=None):
        from lace.meshviewer import MeshViewer

        if mv is None:
            mv = MeshViewer(keepalive=True)

        mv.set_dynamic_lines([self])

        return mv

    def all_edges_with_verts(self, v_indices, as_boolean=False):
        '''
        returns all of the faces that contain at least one of the vertices in v_indices
        '''
        included_vertices = np.zeros(self.v.shape[0], dtype=bool)
        included_vertices[v_indices] = True
        edges_with_verts = included_vertices[self.e].all(axis=1)
        if as_boolean:
            return edges_with_verts
        return np.nonzero(edges_with_verts)[0]

    def keep_vertices(self, indices_to_keep, ret_kept_edges=False):
        '''
        Keep the given vertices and discard the others, and any edges to which
        they may belong.


        If `ret_kept_edges` is `True`, return the original indices of the kept
        edges. Otherwise return `self` for chaining.

        '''

        if self.v is None:
            return

        initial_num_verts = self.v.shape[0]
        if self.e is not None:
            initial_num_edges = self.e.shape[0]
            e_indices_to_keep = self.all_edges_with_verts(indices_to_keep, as_boolean=True)

        self.v = self.v[indices_to_keep]
        if self.e is not None:
            v_old_to_new = np.zeros(initial_num_verts, dtype=int)
            e_old_to_new = np.zeros(initial_num_edges, dtype=int)

            v_old_to_new[indices_to_keep] = np.arange(len(indices_to_keep), dtype=int)
            self.e = v_old_to_new[self.e[e_indices_to_keep]]
            e_old_to_new[e_indices_to_keep] = np.arange(self.e.shape[0], dtype=int)
        else:
            e_indices_to_keep = []

        return np.nonzero(e_indices_to_keep)[0] if ret_kept_edges else self

    def remove_vertices(self, v_list):
        return self.keep_vertices(np.setdiff1d(np.arange(self.v.shape[0]), v_list))



# pylint: disable=attribute-defined-outside-init
import numpy as np
from blmath.numerics import vx

class MeshMixin(object):
    def estimate_vertex_normals(self, face_to_verts_sparse_matrix=None):
        from blmath.optimization.objectives.normals import TriNormalsScaled
        face_normals = TriNormalsScaled(self.v, self.f).r.reshape(-1, 3)
        ftov = face_to_verts_sparse_matrix if face_to_verts_sparse_matrix else self.faces_by_vertex(as_sparse_matrix=True)
        non_scaled_normals = ftov*face_normals
        norms = (np.sum(non_scaled_normals**2.0, axis=1)**0.5).T
        norms[norms == 0] = 1.0
        self.vn = (non_scaled_normals.T/norms).T
        return self.vn # for backwards compatibility

    def barycentric_coordinates_for_points(self, points, face_indices):
        from blmath.geometry.barycentric import barycentric_coordinates_of_projection
        vertex_indices = self.f[face_indices.flatten(), :]
        tri_vertices = np.array([self.v[vertex_indices[:, 0]], self.v[vertex_indices[:, 1]], self.v[vertex_indices[:, 2]]])
        return vertex_indices, barycentric_coordinates_of_projection(points, tri_vertices[0, :], tri_vertices[1, :] - tri_vertices[0, :], tri_vertices[2, :] - tri_vertices[0, :])

    def reset_normals(self, face_to_verts_sparse_matrix=None):
        self.vn = self.estimate_vertex_normals(face_to_verts_sparse_matrix)

    def reset_face_normals(self):
        if not hasattr(self, 'vn'):
            self.reset_normals()
        self.fn = self.f

    def scale(self, scale_factor):
        if self.v is not None:
            self.v *= scale_factor

    def convert_units(self, from_units, to_units):
        '''
        Convert the mesh from one set of units to another.

        These calls are equivalent:

        - mesh.convert_units(from_units='cm', to_units='m')
        - mesh.scale(.01)

        '''
        from blmath import units
        factor = units.factor(
            from_units=from_units,
            to_units=to_units,
            units_class='length'
        )
        self.scale(factor)

    def predict_body_units(self):
        '''
        There is no prediction for united states unit system.
        This may fail when a mesh is not axis-aligned
        '''
        longest_dist = np.max(np.max(self.v, axis=0) - np.min(self.v, axis=0))
        if round(longest_dist / 1000) > 0:
            return 'mm'
        if round(longest_dist / 100) > 0:
            return 'cm'
        if round(longest_dist / 10) > 0:
            return 'dm'
        return 'm'

    def assert_body_units(self, expected):
        if not self.predict_body_units() == expected:
            raise ValueError('Expected body to be in {} but it appears to be in {}')

    def rotate(self, rotation_matrix):
        if np.array(rotation_matrix).shape != (3, 3):
            import cv2
            rotation_matrix = cv2.Rodrigues(np.array(rotation_matrix))[0]
        self.v = np.dot(self.v, rotation_matrix.T)

    def translate(self, translation):
        if self.v is not None:
            self.v += translation

    def reorient(self, up, look):
        '''
        Reorient the mesh by specifying two vectors.

        up: The foot-to-head direction.
        look: The direction the body is facing.

        In the result, the up will end up along +y, and look along +z
        (i.e. facing towards a default OpenGL camera).

        '''
        from blmath.geometry.transform import rotation_from_up_and_look
        from blmath.numerics import as_numeric_array

        up = as_numeric_array(up, (3,))
        look = as_numeric_array(look, (3,))

        if self.v is not None:
            self.v = np.dot(rotation_from_up_and_look(up, look), self.v.T).T

    def flip(self, axis=0, preserve_centroid=False):
        '''
        Flip the mesh across the given axis: 0 for x, 1 for y, 2 for z.

        When `preserve_centroid` is True, translate after flipping to
        preserve the location of the centroid.

        '''
        self.v[:, axis] *= -1

        if preserve_centroid:
            self.v[:, axis] -= 2 * self.centroid[0]

        self.flip_faces()

    # TODO : many of the methods below (e.g. centroid, floor_point, bounding_box)
    # are highly sensative to outliers. A nice compliment to these methods would
    # be a way to copy/modify the mesh so some percentile of outlier are removed.
    @property
    def centroid(self):
        '''
        Return the geometric center.

        '''
        if self.v is None:
            raise ValueError('Mesh has no vertices; centroid is not defined')

        return np.mean(self.v, axis=0)

    def recenter_on_centroid(self):
        self.translate(-self.centroid)

    @property
    def floor_point(self):
        '''
        Return the point on the floor that lies below the centroid.

        '''
        floor_point = self.centroid
        # y to floor
        floor_point[1] = self.v[:, 1].min()
        return floor_point

    @property
    def floor_plane(self):
        from blmath.geometry import Plane
        return Plane(self.floor_point, vx.basis.y)

    def recenter_over_floor(self):
        self.translate(-self.floor_point)

    @property
    def bounding_box(self):
        from blmath.geometry import Box

        if self.v is None:
            raise ValueError('Mesh has no vertices; bounding box is not defined')

        return Box(np.min(self.v, axis=0), np.ptp(self.v, axis=0))

    def almost_on_floor(self, atol=1e-08):
        if self.v is None:
            raise ValueError('Mesh has no vertices; floor is not defined')

        min_y = np.min(self.v, axis=0)[1]
        return np.isclose(min_y, 0., rtol=0, atol=atol)

    def apex(self, axis):
        '''
        Find the most extreme vertex in the direction of the axis provided.

        axis: A vector, which is an 3x1 np.array.

        '''
        from blmath.geometry.apex import apex
        return apex(self.v, axis)

    def first_blip(self, squash_axis, origin, initial_direction):
        '''
        Flatten the mesh to the plane, dropping the dimension identified by
        `squash_axis`: 0 for x, 1 for y, 2 for z. Cast a ray from `origin`,
        pointing along `initial_direction`. Sweep the ray, like a radar, until
        encountering the mesh, and return that vertex: like the first blip of
        the radar. The radar works as if on a point cloud: it sees sees only
        vertices, not edges.

        The ray sweeps clockwise and counterclockwise at the same time, and
        returns the first point it hits.

        If no intersection occurs within 90 degrees, return None.

        `initial_direction` need not be normalized.

        '''
        from blmath.numerics import as_numeric_array

        origin = vx.reject_axis(as_numeric_array(origin, (3,)), axis=squash_axis, squash=True)
        initial_direction = vx.reject_axis(as_numeric_array(initial_direction, (3,)), axis=squash_axis, squash=True)
        vertices = vx.reject_axis(self.v, axis=squash_axis, squash=True)

        origin_to_mesh = vx.normalize(vertices - origin)
        cosines = vx.normalize(initial_direction).dot(origin_to_mesh.T).T
        index_of_first_blip = np.argmax(cosines)

        return self.v[index_of_first_blip]

    def cut_across_axis(self, dim, minval=None, maxval=None):
        '''
        Cut the mesh by a plane, discarding vertices that lie behind that
        plane. Or cut the mesh by two parallel planes, discarding vertices
        that lie outside them.

        The region to keep is defined by an axis of perpendicularity,
        specified by `dim`: 0 means x, 1 means y, 2 means z. `minval`
        and `maxval` indicate the portion of that axis to keep.

        Return the original indices of the kept vertices.

        '''
        # vertex_mask keeps track of the vertices we want to keep.
        vertex_mask = np.ones((len(self.v),), dtype=bool)

        if minval is not None:
            predicate = self.v[:, dim] >= minval
            vertex_mask = np.logical_and(vertex_mask, predicate)

        if maxval is not None:
            predicate = self.v[:, dim] <= maxval
            vertex_mask = np.logical_and(vertex_mask, predicate)

        vertex_indices = np.flatnonzero(vertex_mask)
        self.keep_vertices(vertex_indices)

        return vertex_indices

    def cut_across_axis_by_percentile(self, dim, minpct=0, maxpct=100):
        '''
        Like cut_across_axis, except the subset of vertices is
        constrained by percentile of the data along a given axis
        instead of specific values. (See numpy.percentile.)

        For example, if the mesh has 50,000 vertices, `dim` is 2, and
        `minpct` is 10, this method drops the 5,000 vertices which are
        furthest along the +z axis.

        See numpy.percentile

        Return the original indices of the kept vertices.

        '''
        value_range = np.percentile(self.v[:, dim], (minpct, maxpct))
        return self.cut_across_axis(dim, *value_range)

    def cut_by_plane(self, plane, inverted=False):
        '''
        Like cut_across_axis, but works with an arbitrary plane. Keeps
        vertices that lie in front of the plane (i.e. in the direction
        of the plane normal).

        inverted: When `True`, invert the logic, to keep the vertices
          that lie behind the plane instead.

        Return the original indices of the kept vertices.

        '''
        vertices_to_keep = plane.points_in_front(self.v, inverted=inverted, ret_indices=True)

        self.keep_vertices(vertices_to_keep)

        return vertices_to_keep

    @property
    def surface_area(self):
        '''
        returns the surface area of the mesh
        '''
        return self.surface_areas.sum()

    @property
    def surface_areas(self):
        '''
        returns the surface area of each face
        '''
        e_1 = self.v[self.f[:, 1]] - self.v[self.f[:, 0]]
        e_2 = self.v[self.f[:, 2]] - self.v[self.f[:, 0]]

        cross_products = np.array([e_1[:, 1]*e_2[:, 2] - e_1[:, 2]*e_2[:, 1],
                                   e_1[:, 2]*e_2[:, 0] - e_1[:, 0]*e_2[:, 2],
                                   e_1[:, 0]*e_2[:, 1] - e_1[:, 1]*e_2[:, 0]]).T

        return (0.5)*((cross_products**2.).sum(axis=1)**0.5)



'''
Reads the CAESAR .lmrk file format.

'''

from __future__ import absolute_import
__all__ = ['load', 'loads']

def load(f, *args, **kwargs):
    from baiji.serialization.util.openlib import ensure_file_open_and_call
    return ensure_file_open_and_call(f, _load, 'rb', *args, **kwargs)

def _load(f, *args, **kwargs): # pylint: disable=unused-argument
    return loads(f.read())

def loads(s, ret_commands=False):
    import numpy as np

    landmarks = {}
    commands = {}

    for line in s.splitlines():
        if not line.strip():
            continue

        key = line.split()[0]
        data = [float(x) for x in line.split()[1:]]

        if key == '_scale':
            commands['scale'] = np.matrix(data)
        elif key == '_translate':
            commands['translate'] = np.matrix(data)
        elif key == '_rotation':
            commands['rotation'] = np.matrix(data).reshape(3, 3)
        else:
            landmarks[key] = [data[1], data[2], data[0]]

    return (landmarks, commands) if ret_commands else landmarks



__all__ = ['load', 'dump', 'EXTENSION']

EXTENSION = '.stl'


def load(f, existing_mesh=None):
    from baiji.serialization.util.openlib import ensure_file_open_and_call
    return ensure_file_open_and_call(f, _load, mode='rb', mesh=existing_mesh)


def dump(mesh, f, name="mesh", ascii=True):
    from baiji.serialization.util.openlib import ensure_file_open_and_call
    return ensure_file_open_and_call(f, _dump, mode='wb', mesh=mesh, name=name, ascii=ascii)


def _load(f, mesh=None):
    import numpy as np
    from lace.mesh import Mesh
    if not mesh:
        mesh = Mesh()
    faces = []
    facenormals = []
    verts = []

    head = f.readline().strip()
    if head.startswith("solid"): # ascii STL format
        #name = head[6:]
        current_face = []
        for line in f:
            line = line.split()
            if line[0] == "endsolid":
                break
            elif line[0] == "facet":
                current_face = []
                if line[1] == "normal":
                    try:
                        facenormals.append([float(x) for x in line[2:]])
                    except: # pylint: disable=bare-except
                        facenormals.append([np.nan, np.nan, np.nan])
                else:
                    facenormals.append([np.nan, np.nan, np.nan])
            elif line[0] == "endfacet":
                faces.append(current_face)
                current_face = []
            elif line[0:2] == ["outer", "loop"]:
                pass
            elif line[0] == "endloop":
                pass
            elif line[0] == "vertex":
                current_face.append(len(verts))
                try:
                    verts.append([float(x) for x in line[1:]])
                except: # pylint: disable=bare-except
                    verts.append([np.nan, np.nan, np.nan])
            else:
                raise ValueError("Badly formatted STL file. I don't understand the line %s" % line)
    else:
        raise Exception("Looks like this is a binary STL file; you're going to have to implement that")
        # format docs are here: http://en.wikipedia.org/wiki/STL_(file_format)

    mesh.v = np.array(verts, dtype=np.float64).copy()
    mesh.f = np.array(faces, dtype=np.uint32).copy()
    mesh.fn = np.array(facenormals, dtype=np.float64).copy()
    return mesh


def _dump(f, mesh, name="mesh", ascii=True): # pylint: disable=unused-argument
    raise NotImplementedError("stl.dump is not implemented yet")



import unittest
import numpy as np
from scratch_dir import ScratchDirMixin
from bltest import skip_on_import_error
from bltest.extra_asserts import ExtraAssertionsMixin
from lace.cache import sc
from lace.mesh import Mesh

class TestLandmarks(ExtraAssertionsMixin, ScratchDirMixin, unittest.TestCase):
    def setUp(self):
        skip_on_import_error('lace-search')
        self.scan_fname = sc('s3://bodylabs-korper-assets/is/ps/shared/data/body/caesar/RawScans/csr0001a.ply')
        self.scan_lmrk = sc('s3://bodylabs-korper-assets/is/ps/shared/data/body/caesar/Landmarks/csr0001a.lmrk')
        self.template_fname = sc('s3://bodylabs-korper-assets/is/ps/shared/data/body/template/textured_mean_scape_female.obj')
        self.template_pp = sc('s3://bodylabs-korper-assets/is/ps/shared/data/body/template/template_caesar_picked_points.pp')
        self.scan = Mesh(filename=self.scan_fname, lmrkfilename=self.scan_lmrk)
        self.template = Mesh(filename=self.template_fname, ppfilename=self.template_pp)
        self.template_without_regressors = Mesh(filename=self.template_fname, ppfilename=self.template_pp)
        self.template_without_regressors.landm_regressors = {}
        super(TestLandmarks, self).setUp()

    def test_CAESAR_lmrk_file(self):
        m = Mesh(filename=self.scan_fname, landmarks=self.scan_lmrk)
        self.assertEqual(m.landm, self.scan.landm)
        self.assertDictOfArraysEqual(m.landm_xyz, self.scan.landm_xyz)

    def test_meshlab_pp_file(self):
        m = Mesh(filename=self.template_fname, landmarks=self.template_pp)
        self.assertEqual(m.landm, self.template.landm)
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, self.template.landm_xyz)

    def make_index_data(self, loader, extension):
        import os
        path = os.path.join(self.scratch_dir, 'landmarks' + extension)
        test_data = {name: long(val) for name, val in self.template.landm.items()}
        loader.dump(test_data, path)
        return path

    def make_point_data(self, loader, extension):
        import os
        path = os.path.join(self.scratch_dir, 'landmarks' + extension)
        test_data = {name: val.tolist() for name, val in self.template.landm_xyz.items()}
        loader.dump(test_data, path)
        return path

    def test_yaml_index(self):
        from baiji.serialization import yaml
        path = self.make_index_data(yaml, '.yaml')
        m = Mesh(filename=self.template_fname, landmarks=path)
        self.assertEqual(m.landm, self.template.landm)
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, self.template_without_regressors.landm_xyz)

    def test_yaml_points(self):
        from baiji.serialization import yaml
        path = self.make_point_data(yaml, '.yaml')
        m = Mesh(filename=self.template_fname, landmarks=path)
        self.assertEqual(m.landm, self.template.landm)
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, self.template.landm_xyz)

    def test_yml_index(self):
        from baiji.serialization import yaml
        path = self.make_index_data(yaml, '.yml')
        m = Mesh(filename=self.template_fname, landmarks=path)
        self.assertEqual(m.landm, self.template.landm)
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, self.template_without_regressors.landm_xyz)

    def test_yml_points(self):
        from baiji.serialization import yaml
        path = self.make_point_data(yaml, '.yml')
        m = Mesh(filename=self.template_fname, landmarks=path)
        self.assertEqual(m.landm, self.template.landm)
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, self.template.landm_xyz)

    def test_json_index(self):
        from baiji.serialization import json
        path = self.make_index_data(json, '.json')
        m = Mesh(filename=self.template_fname, landmarks=path)
        self.assertEqual(m.landm, self.template.landm)
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, self.template_without_regressors.landm_xyz)

    def test_json_points(self):
        from baiji.serialization import json
        path = self.make_point_data(json, '.json')
        m = Mesh(filename=self.template_fname, landmarks=path)
        self.assertEqual(m.landm, self.template.landm)
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, self.template.landm_xyz)

    def test_pickle_index(self):
        from baiji.serialization import pickle
        path = self.make_index_data(pickle, '.pkl')
        m = Mesh(filename=self.template_fname, landmarks=path)
        self.assertEqual(m.landm, self.template.landm)
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, self.template_without_regressors.landm_xyz)

    def test_pickle_points(self):
        from baiji.serialization import pickle
        path = self.make_point_data(pickle, '.pkl')
        m = Mesh(filename=self.template_fname, landmarks=path)
        self.assertEqual(m.landm, self.template.landm)
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, self.template.landm_xyz)

    def test_explicit_landmark_indexes_from_another_mesh(self):
        m = Mesh(filename=self.template_fname, landmarks=self.template.landm)
        np.testing.assert_array_equal(self.template.v, m.v)
        self.assertEqual(m.landm, self.template.landm)
        # The following are _not_ expected to pass, as the initial load of template is
        # loading points that are slightly off from the actual vertex locations,
        # so the regressors are going to be subtly different.
        #   self.assertDictOfArraysAlmostEqual(m.landm_xyz, self.template.landm_xyz)
        #   self.assertDictOfArraysAlmostEqual(m.landm_regressors, self.template.landm_regressors)
        # But if we disable the regressors and just use the landmark verts:
        m.landm_regressors = {}
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, self.template_without_regressors.landm_xyz)

    def test_explicit_landmark_points_from_another_mesh(self):
        m = Mesh(filename=self.template_fname, landmarks=self.template.landm_xyz)
        self.assertEqual(m.landm, self.template.landm)
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, self.template.landm_xyz)

    def test_explicit_landmarks_as_list_of_indexes(self):
        m = Mesh(filename=self.template_fname, landmarks=[0, 7])
        self.assertEqual(m.landm, {'0': 0, '1': 7})
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, {'0': self.template.v[0], '1': self.template.v[7]})

    def test_explicit_landmarks_as_array_of_indexes(self):
        m = Mesh(filename=self.template_fname, landmarks=np.array([0, 7]))
        self.assertEqual(m.landm, {'0': 0, '1': 7})
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, {'0': self.template.v[0], '1': self.template.v[7]})

    def test_explicit_landmarks_as_dict_of_indexes(self):
        m = Mesh(filename=self.template_fname, landmarks={'foo': 0, 'bar': 7})
        self.assertEqual(m.landm, {'foo': 0, 'bar': 7})
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, {'foo': self.template.v[0], 'bar': self.template.v[7]})

    def test_explicit_landmarks_as_array_of_points(self):
        m = Mesh(filename=self.template_fname, landmarks=[self.template.v[0], self.template.v[7]])
        self.assertEqual(m.landm, {'0': 0, '1': 7})
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, {'0': self.template.v[0], '1': self.template.v[7]})

    def test_explicit_landmarks_as_list_of_points(self):
        m = Mesh(filename=self.template_fname, landmarks=[self.template.v[0].tolist(), self.template.v[7].tolist()])
        self.assertEqual(m.landm, {'0': 0, '1': 7})
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, {'0': self.template.v[0], '1': self.template.v[7]})

    def test_explicit_landmarks_as_dict_of_points(self):
        m = Mesh(filename=self.template_fname, landmarks={'foo': self.template.v[0], 'bar': self.template.v[7]})
        self.assertEqual(m.landm, {'foo': 0, 'bar': 7})
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, {'foo': self.template.v[0], 'bar': self.template.v[7]})

    def test_explicit_landmarks_as_array_of_close_points(self):
        m = Mesh(filename=self.template_fname, landmarks=[self.template.v[0]+.000001, self.template.v[7]+.000001])
        self.assertEqual(m.landm, {'0': 0, '1': 7})
        self.assertDictOfArraysAlmostEqual(m.landm_xyz, {'0': self.template.v[0], '1': self.template.v[7]})



import unittest
import copy
import numpy as np
from lace.arcball import Matrix4fT, Matrix3fT, ArcBallT, Point2fT, Matrix3fSetRotationFromQuat4f, Matrix3fMulMatrix3f, Matrix4fSetRotationFromMatrix3f

class TestArcBall(unittest.TestCase):
    def test_ArcBall(self):
        # Unit testing of the ArcBall calss and the real math behind it.
        # Simulates a click and drag followed by another click and drag.
        Transform = Matrix4fT()
        LastRot = Matrix3fT()
        ThisRot = Matrix3fT()

        ArcBall = ArcBallT(640, 480)

        # First click
        LastRot = copy.copy(ThisRot)
        mouse_pt = Point2fT(500, 250)
        ArcBall.click(mouse_pt)
        # First drag
        mouse_pt = Point2fT(475, 275)
        ThisQuat = ArcBall.drag(mouse_pt)
        np.testing.assert_array_almost_equal(ThisQuat, [0.08438914, -0.08534209, -0.06240178, 0.99080837])
        ThisRot = Matrix3fSetRotationFromQuat4f(ThisQuat)
        # Linear Algebra matrix multiplication A = old, B = New : C = A * B
        ThisRot = Matrix3fMulMatrix3f(LastRot, ThisRot)
        Transform = Matrix4fSetRotationFromMatrix3f(Transform, ThisRot)
        np.testing.assert_array_almost_equal(Transform,
                                             [[0.97764552, -0.1380603, 0.15858325, 0.],
                                              [0.10925253, 0.97796899, 0.17787792, 0.],
                                              [-0.17964739, -0.15657592, 0.97119039, 0.],
                                              [0., 0., 0., 1.]])
        # Done with first drag

        # second click
        LastRot = copy.copy(ThisRot)
        np.testing.assert_array_almost_equal(LastRot,
                                             [[0.97764552, -0.1380603, 0.15858325],
                                              [0.10925253, 0.97796899, 0.17787792],
                                              [-0.17964739, -0.15657592, 0.97119039]])
        mouse_pt = Point2fT(350, 260)
        ArcBall.click(mouse_pt)
        # second drag
        mouse_pt = Point2fT(450, 260)
        ThisQuat = ArcBall.drag(mouse_pt)
        np.testing.assert_array_almost_equal(ThisQuat, [0.00710336, 0.31832787, 0.02679029, 0.94757545])
        ThisRot = Matrix3fSetRotationFromQuat4f(ThisQuat)
        ThisRot = Matrix3fMulMatrix3f(LastRot, ThisRot)
        Transform = Matrix4fSetRotationFromMatrix3f(Transform, ThisRot)
        np.testing.assert_array_almost_equal(Transform,
                                             [[0.88022292, -0.08322023, -0.46720669, 0.],
                                              [0.14910145, 0.98314685, 0.10578787, 0.],
                                              [0.45052907, -0.16277808, 0.8777966, 0.],
                                              [0., 0., 0., 1.00000001]])
        # Done with second drag
        LastRot = copy.copy(ThisRot)


if __name__ == '__main__': # pragma: no cover
    unittest.main()



# pylint: disable=len-as-condition
__all__ = ['load', 'ParseError']

class ParseError(Exception):
    pass

def load(f, existing_mesh=None):
    from baiji.serialization.util.openlib import ensure_file_open_and_call
    return ensure_file_open_and_call(f, _load, mode='rb', mesh=existing_mesh)

def _load(f, mesh=None):
    from lace.mesh import Mesh

    try:
        v, f, vn, vc = _parse_wrl(f)
    except Exception:
        import traceback
        tb = traceback.format_exc()
        raise ParseError("Unable to parse wrl file with exception : \n[\n%s]" % tb)

    if not mesh:
        mesh = Mesh()
    if v.size != 0:
        mesh.v = v
    if f.size != 0:
        mesh.f = f
    if vn.size != 0:
        mesh.vn = vn
    if vc.size != 0:
        mesh.vc = vc
    return mesh

def _find_next_block(f, keywords=['coord', 'color', 'normal', 'coordIndex']):
    for line in f:
        line_splitted = line.split()
        if not line_splitted:
            continue
        if line_splitted[0] in keywords:
            return line_splitted[0]

    return None

def _load_faces(f, target, start_index):
    for face_line in f:
        if not face_line.strip():
            continue
        i = face_line.strip().find('-1')
        if i < 0:
            break
        target.append(map(lambda x: int(x) + start_index, face_line.split(',')[:3]))

def _load_per_vertex_data(f, target):
    for vertex_line in f:

        if not vertex_line.strip():
            continue

        i = vertex_line.strip().find(',')
        if i < 0:
            break
        target.append(map(float, vertex_line.strip()[:i].split()))

def _parse_wrl(f):

    import numpy as np
    v = []
    vn = []
    vc = []
    faces = []
    start_index = 0

    format_disclaimer = f.readline()
    if format_disclaimer.strip() != '#VRML V2.0 utf8':
        raise ParseError("Unsupported wrl file format")

    while True:
        key = _find_next_block(f)
        if key == 'coord':
            _load_per_vertex_data(f, v)
        elif key == 'color':
            _load_per_vertex_data(f, vc)
        elif key == 'normal':
            _load_per_vertex_data(f, vn)
        elif key == 'coordIndex':
            _load_faces(f, faces, start_index)
            start_index = len(v)
        else:
            break

    if len(v) == 0:
        raise ParseError("Unsupported wrl file format")
    return np.array(v), np.array(faces), np.array(vn), np.array(vc)



import unittest

sample_xml_string = """
<!DOCTYPE PickedPoints>
<PickedPoints>
 <DocumentData>
  <DateTime time="16:00:00" date="2014-12-31"/>
  <User name="bodylabs"/>
  <DataFileName name="a_pose.ply"/>
 </DocumentData>
 <point x="0.044259" y="0.467733" z="-0.060032" name="Femoral_epicon_med_lft"/>
<point x="0.017893" y="1.335375" z="0.018390" name="Clavicale_lft"/>
<point x="0.000625" y="1.124424" z="0.080930" name="Substernale"/>
</PickedPoints>
"""

sample_points = {
    'Femoral_epicon_med_lft': [0.044259, 0.467733, -0.060032],
    'Clavicale_lft': [0.017893, 1.335375, 0.018390],
    'Substernale': [0.000625, 1.124424, 0.080930],
}

class TestMeshlabPickedPoints(unittest.TestCase):

    def assertEqualPickedPointXml(self, value, expected):
        # This is fairly naive -- doesn't consider whitespace or that the points
        # could be in different order, and is strict about irrelevant parts like
        #the DateTime. But seems to work okay.
        from lxml import etree, objectify

        value_normalized = etree.tostring(objectify.fromstring(value))
        expected_normalized = etree.tostring(objectify.fromstring(expected))

        self.assertEquals(value_normalized, expected_normalized)

    def test_load(self):
        import StringIO
        from lace.serialization import meshlab_pickedpoints

        sample_f = StringIO.StringIO(sample_xml_string)

        try:
            result = meshlab_pickedpoints.load(sample_f)
        finally:
            sample_f.close()

        self.assertEqual(result, sample_points)

    def test_dump(self):
        import StringIO
        from lace.serialization import meshlab_pickedpoints

        result_f = StringIO.StringIO()

        try:
            meshlab_pickedpoints.dump(sample_points, result_f, mesh_filename='a_pose.ply')
            result_str = result_f.getvalue()
        finally:
            result_f.close()

        self.assertEqualPickedPointXml(result_str, sample_xml_string)


if __name__ == '__main__': # pragma: no cover
    unittest.main()



class MeshMixin(object):
    def vertex_visibility(self, camera, normal_threshold=None, omni_directional_camera=False, binary_visiblity=True):
        import numpy as np
        vis, n_dot_cam = self.vertex_visibility_and_normals(camera, omni_directional_camera)
        if normal_threshold is not None:
            vis = np.logical_and(vis, n_dot_cam > normal_threshold)
        return np.squeeze(vis) if binary_visiblity else np.squeeze(vis*n_dot_cam)

    def vertex_visibility_and_normals(self, camera, omni_directional_camera=False):
        import numpy as np
        from lace_search.visibility import visibility_compute # pylint: disable=no-name-in-module
        arguments = {'v': self.v, 'f': self.f, 'cams' : np.array([camera.origin.flatten()])}
        if not omni_directional_camera:
            arguments['sensors'] = np.array([camera.sensor_axis.flatten()])
        arguments['n'] = self.vn if hasattr(self, 'vn') else self.estimate_vertex_normals()

        return visibility_compute(**arguments)

    def visibile_mesh(self, camera=[0.0, 0.0, 0.0]):
        import numpy as np
        vis = self.vertex_visibility(camera)
        faces_to_keep = filter(lambda face: vis[face[0]]*vis[face[1]]*vis[face[2]], self.f)
        vertex_indices_to_keep = np.nonzero(vis)[0]
        vertices_to_keep = self.v[vertex_indices_to_keep]
        old_to_new_indices = np.zeros(len(vis))
        old_to_new_indices[vertex_indices_to_keep] = range(len(vertex_indices_to_keep))
        return self.__class__(v=vertices_to_keep, f=np.array([old_to_new_indices[face] for face in faces_to_keep]))



import unittest
import numpy as np
from lace.cache import sc
from lace.serialization import stl

class TestSTL(unittest.TestCase):
    def setUp(self):
        import tempfile
        self.tmp_dir = tempfile.mkdtemp('bodylabs-test')
        self.truth = {
            'box_v': np.array([[0.5, -0.5, 0.5, -0.5, 0.5, -0.5, 0.5, -0.5], [0.5, 0.5, -0.5, -0.5, 0.5, 0.5, -0.5, -0.5], [0.5, 0.5, 0.5, 0.5, -0.5, -0.5, -0.5, -0.5]]).T,
            'box_f': np.array([[0, 1, 2], [3, 2, 1], [0, 2, 4], [6, 4, 2], [0, 4, 1], [5, 1, 4], [7, 5, 6], [4, 6, 5], [7, 6, 3], [2, 3, 6], [7, 3, 5], [1, 5, 3]]),
            'box_fn': np.array([[0, 0, 1], [0, 0, 1], [1, 0, 0], [1, 0, 0], [0, 1, 0], [0, 1, 0], [0, 0, -1], [-0, -0, -1], [0, -1, 0], [-0, -1, -0], [-1, 0, 0], [-1, -0, -0]]),
        }
        # Because STL gives duplicated verts
        self.truth['box_v'] = self.truth['box_v'][self.truth['box_f'].flatten()]
        self.truth['box_f'] = np.array(range(self.truth['box_v'].shape[0])).reshape((-1, 3))
        self.test_stl_url = "s3://bodylabs-korper-assets/is/ps/shared/data/body/korper_testdata/test_box.stl"
        self.test_stl_path = sc(self.test_stl_url)

    def tearDown(self):
        import shutil
        shutil.rmtree(self.tmp_dir, ignore_errors=True)

    def test_loads_from_local_path_using_serializer(self):
        m = stl.load(self.test_stl_path)
        np.testing.assert_array_equal(m.v, self.truth['box_v'])
        np.testing.assert_array_equal(m.f, self.truth['box_f'])
        np.testing.assert_array_equal(m.fn, self.truth['box_fn'])



#!/usr/bin/env python
# pylint: disable=len-as-condition, attribute-defined-outside-init, len-as-condition, invalid-unary-operand-type
# encoding: utf-8

import sys
import os
import os.path
import time
import copy
import subprocess
import re
import platform
from multiprocessing import freeze_support
import numpy as np
import OpenGL.GL as gl
from OpenGL.GL import shaders
import OpenGL.GLUT as glut
import OpenGL.GLU as glu
import OpenGL.arrays.vbo
from blmath.numerics.matlab import row
from blmath.optimization.objectives.normals import TriNormals
import zmq
from lace import arcball
from lace.mesh import Mesh

__all__ = ['MeshViewer', 'MeshViewers', 'test_for_opengl']

"""
MeshViewer.py

Created by Matthew Loper on 2012-05-11.
Copyright (c) 2012 MPI. All rights reserved.
"""


def _popen_exec_python(command, args=[], stdin=None, stdout=None, stderr=None):
    if platform.system() == 'Windows':
        python_executable = ["C:\\Python27\\python.exe"]
    else:
        python_executable = ["/usr/bin/env", "python"]
    return subprocess.Popen(python_executable + [command] + args, stdin=stdin, stdout=stdout, stderr=stderr)


def _run_self(args, stdin=None, stdout=subprocess.PIPE, stderr=None):
    if platform.system() == 'Windows':
        try:
            from py_windows_exe import running_as_windows_exe, get_main_dir
            if running_as_windows_exe():
                env = os.environ.copy()
                env["PATH"] = "%s;%s" % (get_main_dir(), env["PATH"])
                return subprocess.Popen([os.path.join(get_main_dir(), 'pymeshviewer.exe')] + args, stdin=stdin, stdout=stdout, stderr=stderr, env=env)
        except Exception: # pylint: disable=broad-except
            pass
    return _popen_exec_python(os.path.abspath(__file__), args, stdin=stdin, stdout=stdout, stderr=stderr)


def test_for_opengl():
    if test_for_opengl.result is None:
        p = _run_self(["TEST_FOR_OPENGL"], stderr=open(os.devnull, 'wb'))
        line = p.stdout.readline()
        test_for_opengl.result = 'success' in line
    return test_for_opengl.result
test_for_opengl.result = None

class Dummy(object):
    def __getattr__(self, name):
        return Dummy()

    def __call__(self, *args, **kwargs):
        return Dummy()

    def __getitem__(self, key):
        return Dummy()

    def __setitem__(self, key, value):
        pass


def MeshViewer(titlebar='Mesh Viewer', static_meshes=None, static_lines=None, uid=None, autorecenter=True, keepalive=False, window_width=1280, window_height=960, snapshot_camera=None):
    """Allows visual inspection of geometric primitives.

    Write-only Attributes:
        titlebar: string printed in the window titlebar
        dynamic_meshes: list of Mesh objects to be displayed
        static_meshes: list of Mesh objects to be displayed
        dynamic_lines: list of Lines objects to be displayed
        static_lines: list of Lines objects to be displayed

    Note: static_meshes is meant for Meshes that are
    updated infrequently, and dynamic_meshes is for Meshes
    that are updated frequently (same for dynamic_lines vs
    static_lines). They may be treated differently for
    performance reasons.
    """
    if not test_for_opengl():
        return Dummy()
    mv = MeshViewerLocal(shape=(1, 1), uid=uid, titlebar=titlebar, keepalive=keepalive, window_width=window_width, window_height=window_height)
    result = mv.get_subwindows()[0][0]
    result.snapshot_camera = snapshot_camera
    if static_meshes:
        result.static_meshes = static_meshes
    if static_lines:
        result.static_lines = static_lines
    result.autorecenter = autorecenter
    return result


def MeshViewers(shape=(1, 1), titlebar="Mesh Viewers", keepalive=False, window_width=1280, window_height=960):
    """Allows subplot-style inspection of primitives in multiple subwindows.

    Args:
        shape: a tuple indicating the number of vertical and horizontal windows requested

    Returns: a list of lists of MeshViewer objects: one per window requested.
    """
    if not test_for_opengl():
        return Dummy()
    mv = MeshViewerLocal(shape=shape, titlebar=titlebar, uid=None, keepalive=keepalive, window_width=window_width, window_height=window_height)
    return mv.get_subwindows()


class MeshSubwindow(object):
    def __init__(self, parent_window, which_window):
        self.parent_window = parent_window
        self.which_window = which_window

    def set_dynamic_meshes(self, list_of_meshes, blocking=False):
        self.parent_window.set_dynamic_meshes(list_of_meshes, blocking, self.which_window)

    def set_static_meshes(self, list_of_meshes, blocking=False):
        self.parent_window.set_static_meshes(list_of_meshes, blocking, self.which_window)

    # list_of_model_names_and_parameters should be of form [{'name': scape_model_name, 'parameters': scape_model_parameters}]
    # here scape_model_name is the filepath of the scape model.
    def set_dynamic_models(self, list_of_model_names_and_parameters, blocking=False):
        self.parent_window.set_dynamic_models(list_of_model_names_and_parameters, blocking, self.which_window)

    def set_dynamic_lines(self, list_of_lines, blocking=False):
        self.parent_window.set_dynamic_lines(list_of_lines, blocking, self.which_window)

    def set_static_lines(self, list_of_lines, blocking=False):
        self.parent_window.set_static_lines(list_of_lines, blocking=blocking, which_window=self.which_window)

    def set_titlebar(self, titlebar, blocking=False):
        self.parent_window.set_titlebar(titlebar, blocking, which_window=self.which_window)

    def set_autorecenter(self, autorecenter, blocking=False):
        self.parent_window.set_autorecenter(autorecenter, blocking=blocking, which_window=self.which_window)

    def set_background_color(self, background_color, blocking=False):
        self.parent_window.set_background_color(background_color, blocking=blocking, which_window=self.which_window)

    def save_snapshot(self, path, blocking=False):
        self.parent_window.save_snapshot(path, blocking=blocking, which_window=self.which_window)

    def get_event(self):
        return self.parent_window.get_event()

    def get_keypress(self):
        return self.parent_window.get_keypress()['key']

    def get_mouseclick(self):
        return self.parent_window.get_mouseclick()

    def close(self):
        self.parent_window.p.terminate()

    background_color = property(fset=set_background_color, doc="Background color, as 3-element numpy array where 0 <= color <= 1.0.")
    dynamic_meshes = property(fset=set_dynamic_meshes, doc="List of meshes for dynamic display.")
    static_meshes = property(fset=set_static_meshes, doc="List of meshes for static display.")
    dynamic_models = property(fset=set_dynamic_models, doc="List of model names and parameters for dynamic display.")
    dynamic_lines = property(fset=set_dynamic_lines, doc="List of Lines for dynamic display.")
    static_lines = property(fset=set_static_lines, doc="List of Lines for static display.")
    titlebar = property(fset=set_titlebar, doc="Titlebar string.")


class MeshViewerLocal(object):
    """Allows visual inspection of geometric primitives.

    Write-only Attributes:
        titlebar: string printed in the window titlebar
        dynamic_meshes: list of Mesh objects to be displayed
        static_meshes: list of Mesh objects to be displayed
        dynamic_lines: list of Lines objects to be displayed
        static_lines: list of Lines objects to be displayed

    Note: static_meshes is meant for Meshes that are
    updated infrequently, and dynamic_meshes is for Meshes
    that are updated frequently (same for dynamic_lines vs
    static_lines). They may be treated differently for
    performance reasons.
    """
    managed = {}

    def __new__(cls, titlebar, uid, shape, keepalive, window_width, window_height):
        import traceback
        assert uid is None or isinstance(uid, str) or isinstance(uid, unicode)
        if uid == 'stack':
            uid = ''.join(traceback.format_list(traceback.extract_stack()))
        if uid and uid in MeshViewer.managed.keys():
            return MeshViewer.managed[uid]
        result = super(MeshViewerLocal, cls).__new__(cls)
        result.client = zmq.Context.instance().socket(zmq.PUSH)
        result.client.linger = 0
        # result.p = _popen_exec_python([os.path.abspath (inspect.stack()[0][1]), titlebar, str(shape[0]), str(shape[1]), str(window_width), str(window_height)], stdout=subprocess.PIPE)
        result.p = _run_self([titlebar, str(shape[0]), str(shape[1]), str(window_width), str(window_height)])
        line = result.p.stdout.readline()
        current_port = re.match('<PORT>(.*?)</PORT>', line)
        if not current_port:
            raise Exception("MeshViewer remote appears to have failed to launch")
        current_port = int(current_port.group(1))
        result.client.connect('tcp://127.0.0.1:%d' % (current_port))
        if uid:
            MeshViewerLocal.managed[uid] = result
        result.shape = shape
        result.keepalive = keepalive
        return result

    def get_subwindows(self):
        return [[MeshSubwindow(parent_window=self, which_window=(r, c)) for c in range(self.shape[1])] for r in range(self.shape[0])]

    @staticmethod
    def _sanitize_meshes(list_of_meshes):
        lm = []
        # have to copy the meshes for now, because some contain CPython members,
        # before pushing them on the queue
        for m in list_of_meshes:
            #f = m.f if hasattr(m, 'f') and m.f is not None else []
            if m.f is not None:
                f = m.f
                if m.fc is not None:
                    lm.append(Mesh(v=m.v, f=f, fc=m.fc))
                elif m.vc is not None:
                    lm.append(Mesh(v=m.v, f=f, vc=m.vc))
                else:
                    lm.append(Mesh(v=m.v, f=f))
            else:
                if m.fc is not None:
                    lm.append(Mesh(v=m.v, fc=m.fc))
                elif m.vc is not None:
                    lm.append(Mesh(v=m.v, vc=m.vc))
                else:
                    lm.append(Mesh(v=m.v))
            if m.vn is not None:
                lm[-1].vn = m.vn
            if m.fn is not None:
                lm[-1].fn = m.fn
            if m.texture_filepath is not None and m.vt is not None and m.ft is not None:
                lm[-1].texture_filepath = m.texture_filepath
                lm[-1].vt = m.vt
                lm[-1].ft = m.ft
            if m.landm_xyz is not None and m.landm is not None:
                from lace.sphere import Sphere
                sphere = Sphere()
                scalefactor = 1e-2 * np.max(np.max(m.v) - np.min(m.v)) / np.max(np.max(sphere.v) - np.min(sphere.v))
                sphere.scale(scalefactor)
                spheres = [Mesh(vc='SteelBlue', f=sphere.f,
                                v=sphere.v + row(np.array(m.landm_xyz[k])))
                           for k in m.landm.keys()]
                lm.extend(spheres)
        return lm

    def _send_pyobj(self, label, obj, blocking, which_window):
        if blocking:
            context = zmq.Context.instance()
            server = context.socket(zmq.PULL)
            server.linger = 0
            port = server.bind_to_random_port('tcp://127.0.0.1', min_port=49152, max_port=65535, max_tries=100000)
#            print 'sending with blocking'
            self.client.send_pyobj({'label': label, 'obj': obj, 'port': port, 'which_window': which_window})
            task_completion_time = server.recv_pyobj() # pylint: disable=unused-variable
#            print 'task completion time was %.2fs in other process' % (task_completion_time, )
            server.close()
        else:
#            print 'sending nonblocking'
            self.client.send_pyobj({'label': label, 'obj': obj, 'which_window': which_window})

    def set_dynamic_meshes(self, list_of_meshes, blocking=False, which_window=(0, 0)):
        self._send_pyobj('dynamic_meshes', self._sanitize_meshes(list_of_meshes), blocking, which_window)

    def set_static_meshes(self, list_of_meshes, blocking=False, which_window=(0, 0)):
        self._send_pyobj('static_meshes', self._sanitize_meshes(list_of_meshes), blocking, which_window)

    # list_of_model_names_and_parameters should be of form [{'name': scape_model_name, 'parameters': scape_model_parameters}]
    # here scape_model_name is the filepath of the scape model.
    def set_dynamic_models(self, list_of_model_names_and_parameters, blocking=False, which_window=(0, 0)):
        self._send_pyobj('dynamic_models', list_of_model_names_and_parameters, blocking, which_window)

    def set_dynamic_lines(self, list_of_lines, blocking=False, which_window=(0, 0)):
        self._send_pyobj('dynamic_lines', list_of_lines, blocking, which_window)

    def set_static_lines(self, list_of_lines, blocking=False, which_window=(0, 0)):
        self._send_pyobj('static_lines', list_of_lines, blocking, which_window)

    def set_titlebar(self, titlebar, blocking=False, which_window=(0, 0)):
        self._send_pyobj('titlebar', titlebar, blocking, which_window)

    def set_autorecenter(self, autorecenter, blocking=False, which_window=(0, 0)):
        self._send_pyobj('autorecenter', autorecenter, blocking, which_window)

    def set_background_color(self, background_color, blocking=False, which_window=(0, 0)):
        assert isinstance(background_color, np.ndarray)
        assert background_color.size == 3
        self._send_pyobj('background_color', background_color.flatten(), blocking, which_window)

    def get_keypress(self):
        return self.get_ui_event('get_keypress')

    def get_mouseclick(self):
        return self.get_ui_event('get_mouseclick')

    def get_event(self):
        return self.get_ui_event('get_event')

    def get_ui_event(self, event_id):
        context = zmq.Context.instance()
        server = context.socket(zmq.PULL)
        server.linger = 0
        port = server.bind_to_random_port('tcp://127.0.0.1', min_port=49152, max_port=65535, max_tries=100000)
        self._send_pyobj(event_id, port, blocking=True, which_window=(0, 0))
        result = server.recv_pyobj()
        server.close()
        return result

    background_color = property(fset=set_background_color, doc="Background color, as 3-element numpy array where 0 <= color <= 1.0.")
    dynamic_meshes = property(fset=set_dynamic_meshes, doc="List of meshes for dynamic display.")
    static_meshes = property(fset=set_static_meshes, doc="List of meshes for static display.")
    dynamic_models = property(fset=set_dynamic_models, doc="List of model names and parameters for dynamic display.")
    dynamic_lines = property(fset=set_dynamic_lines, doc="List of Lines for dynamic display.")
    static_lines = property(fset=set_static_lines, doc="List of Lines for static display.")
    titlebar = property(fset=set_titlebar, doc="Titlebar string.")

    def save_snapshot(self, path, blocking=False, which_window=(0, 0)):
        self._send_pyobj('save_snapshot', path, blocking, which_window)

    def __del__(self):
        if not self.keepalive:
            self.p.terminate()


class MeshViewerSingle(object):
    '''
    two_sided_lighting: When True, turn on two-sided lighting. I'm deferring the work
      of figuring out how to pass this option e.g. from mesh.show() through the
      MeshViewerRemote and MeshViewerLocal to this object which needs it.

    '''
    def __init__(self, x1_pct, y1_pct, width_pct, height_pct, two_sided_lighting=False):
        assert width_pct <= 1
        assert height_pct <= 1
        self.dynamic_meshes = []
        self.static_meshes = []
        self.dynamic_models = []
        self.dynamic_lines = []
        self.static_lines = []
        self.scape_models = {}
        self.x1_pct = x1_pct
        self.y1_pct = y1_pct
        self.width_pct = width_pct
        self.height_pct = height_pct
        self.autorecenter = True
        self.two_sided_lighting = two_sided_lighting

    def get_dimensions(self):
        d = {}
        d['window_width'] = glut.glutGet(glut.GLUT_WINDOW_WIDTH)
        d['window_height'] = glut.glutGet(glut.GLUT_WINDOW_HEIGHT)
        d['subwindow_width'] = self.width_pct*d['window_width']
        d['subwindow_height'] = self.height_pct*d['window_height']
        d['subwindow_origin_x'] = self.x1_pct*d['window_width']
        d['subwindow_origin_y'] = self.y1_pct*d['window_height']
        return d

    def on_draw(self, transform, want_camera=False):
        d = self.get_dimensions()
        gl.glViewport(
            int(d['subwindow_origin_x']),
            int(d['subwindow_origin_y']),
            int(d['subwindow_width']),
            int(d['subwindow_height']))
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        # Note that the near clip plane is 1 (hither) and the far plane is 1000 (yon)
        glu.gluPerspective(45.0, float(d['subwindow_width'])/float(d['subwindow_height']), 1, 100.0)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()
        if self.two_sided_lighting:
            gl.glLightModeli(gl.GL_LIGHT_MODEL_TWO_SIDE, gl.GL_TRUE)
        else:
            gl.glLightModeli(gl.GL_LIGHT_MODEL_TWO_SIDE, gl.GL_FALSE)
        gl.glTranslatef(0.0, 0.0, -6.0)
        #glTranslatef(0.0, 0.0, -3.5)
        gl.glPushMatrix()
        gl.glMultMatrixf(transform)
        gl.glColor3f(1.0, 0.75, 0.75)
        if self.autorecenter:
            camera = self.draw_primitives_recentered(want_camera=want_camera)
        else:
            if hasattr(self, 'current_center') and hasattr(self, 'current_scalefactor'):
                camera = self.draw_primitives(scalefactor=self.current_scalefactor, center=self.current_center)
            else:
                camera = self.draw_primitives(want_camera=want_camera)
        gl.glPopMatrix()
        if want_camera:
            return camera

    def draw_primitives_recentered(self, want_camera=False):
        return self.draw_primitives(recenter=True, want_camera=want_camera)

    @staticmethod
    def set_shaders(m):
        VERTEX_SHADER = shaders.compileShader("""void main() {
                    gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
                }""", gl.GL_VERTEX_SHADER)
        FRAGMENT_SHADER = shaders.compileShader("""void main() {
                    gl_FragColor = vec4( 0, 1, 0, 1 );
                }""", gl.GL_FRAGMENT_SHADER)
        m.shaders = shaders.compileProgram(VERTEX_SHADER, FRAGMENT_SHADER)

    @staticmethod
    def set_texture(m):
        texture_data = np.array(m.texture_image, dtype='int8')
        m.textureID = gl.glGenTextures(1)
        gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)
        gl.glBindTexture(gl.GL_TEXTURE_2D, m.textureID)
        gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGB, texture_data.shape[1], texture_data.shape[0], 0, gl.GL_BGR, gl.GL_UNSIGNED_BYTE, texture_data.flatten())
        gl.glHint(gl.GL_GENERATE_MIPMAP_HINT, gl.GL_NICEST) # must be GL_FASTEST, GL_NICEST or GL_DONT_CARE
        gl.glGenerateMipmap(gl.GL_TEXTURE_2D)

    @staticmethod
    def draw_mesh(m):
        # Supply vertices
        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
        m.vbo['v'].bind()
        gl.glVertexPointer(3, gl.GL_FLOAT, 0, m.vbo['v'])
        m.vbo['v'].unbind()
        # Supply normals
        if 'vn' in m.vbo.keys():
            gl.glEnableClientState(gl.GL_NORMAL_ARRAY)
            m.vbo['vn'].bind()
            gl.glNormalPointer(gl.GL_FLOAT, 0, m.vbo['vn'])
            m.vbo['vn'].unbind()
        else:
            gl.glDisableClientState(gl.GL_NORMAL_ARRAY)
        # Supply colors
        if 'vc' in m.vbo.keys():
            gl.glEnableClientState(gl.GL_COLOR_ARRAY)
            m.vbo['vc'].bind()
            gl.glColorPointer(3, gl.GL_FLOAT, 0, m.vbo['vc'])
            m.vbo['vc'].unbind()
        else:
            gl.glDisableClientState(gl.GL_COLOR_ARRAY)
        if ('vt' in m.vbo.keys()) and hasattr(m, 'textureID'):
            gl.glEnable(gl.GL_TEXTURE_2D)
            gl.glTexParameterf(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST)
            gl.glTexParameterf(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_NEAREST)
            #GL_DECAL shows texture without additional shading
#            gl.glTexEnvf(gl.GL_TEXTURE_ENV, gl.GL_TEXTURE_ENV_MODE, gl.GL_DECAL)
            gl.glTexEnvf(gl.GL_TEXTURE_ENV, gl.GL_TEXTURE_ENV_MODE, gl.GL_MODULATE)
            gl.glBindTexture(gl.GL_TEXTURE_2D, m.textureID)
            gl.glEnableClientState(gl.GL_TEXTURE_COORD_ARRAY)
            m.vbo['vt'].bind()
            gl.glTexCoordPointer(2, gl.GL_FLOAT, 0, m.vbo['vt'])
            m.vbo['vt'].unbind()
        else:
            gl.glDisable(gl.GL_TEXTURE_2D)
            gl.glDisableClientState(gl.GL_TEXTURE_COORD_ARRAY)
        # Draw
        if m.f is not None and len(m.f) > 0: # ie if it is triangulated
            gl.glEnable(gl.GL_LIGHTING)
            gl.glDrawElementsui(gl.GL_TRIANGLES, np.arange(m.f.size, dtype=np.uint32))
        else: # not triangulated, so disable lighting
            gl.glDisable(gl.GL_LIGHTING)
            gl.glPointSize(2)
            gl.glDrawElementsui(gl.GL_POINTS, np.arange(len(m.v), dtype=np.uint32))

    @staticmethod
    def draw_lines(ls):
        gl.glDisableClientState(gl.GL_NORMAL_ARRAY)
        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
        gl.glLineWidth(1.0)
        allpts = ls.v[ls.e.flatten()].astype(np.float32)
        gl.glVertexPointerf(allpts)
        if ls.vc is not None or ls.ec is not None:
            gl.glEnableClientState(gl.GL_COLOR_ARRAY)
            if ls.vc is not None:
                gl.glColorPointerf(ls.vc[ls.e.flatten()].astype(np.float32))
            else:
                clrs = np.ones((ls.e.shape[0]*2, 3)) * np.repeat(ls.ec, 2, axis=0)
                gl.glColorPointerf(clrs)
        else:
            gl.glDisableClientState(gl.GL_COLOR_ARRAY)
        gl.glDisable(gl.GL_LIGHTING)
        gl.glDrawElementsui(gl.GL_LINES, np.arange(len(allpts), dtype=np.uint32))

    def generate_dynamic_model_meshes(self, remove_head=False):
        try:
            from bodylabs.scape.scapemodel import ScapeModel
        except ImportError:
            raise ImportError("MeshViewer.generate_dynamic_model_meshes requires access to BodyLabs core.")
        for dynamic_model in self.dynamic_models:
            scapemodel_fname = dynamic_model['name']
            if scapemodel_fname not in self.scape_models.keys():
                self.scape_models[scapemodel_fname] = ScapeModel(scapemodel_fname)
                if not hasattr(self.scape_models[dynamic_model['name']].template, 'fbv_cached'):
                    self.scape_models[dynamic_model['name']].template.fbv_cached = self.scape_models[dynamic_model['name']].template.faces_by_vertex()

        model_meshes = []
        for dynamic_model in self.dynamic_models:
            model_parameters = dynamic_model['parameters'].copy()
            auto_translate = ('trans' in model_parameters.keys()) and (model_parameters['trans'] == 'auto')
            if auto_translate:
                del model_parameters['trans']
            model_mesh = self.scape_models[dynamic_model['name']].mesh_for(**model_parameters)
            if auto_translate:
                model_mesh.v = model_mesh.v - np.array([0.0, model_mesh.v[:, 1].min(), 0.0])
                # model_mesh.v = model_mesh.v - np.array([0.0, 0.0, model_mesh.v[:, 2].max()])
            model_mesh.reset_normals(self.scape_models[dynamic_model['name']].template.fbv_cached)
            model_mesh.set_vertex_colors('pink')
            model_meshes.append(model_mesh)
        if remove_head:
            for i, model_mesh in enumerate(model_meshes):
                model_mesh.remove_faces(self.scape_models[self.dynamic_models[i]['name']].template.segm['head'])
        return model_meshes

    def draw_primitives(self, scalefactor=1.0, center=[0.0, 0.0, 0.0], recenter=False, want_camera=False):
        # measure the bounding box of all our primitives, so that we can
        # recenter them in our field of view
        if len(self.dynamic_models):
            all_meshes = self.static_meshes + self.dynamic_meshes + self.generate_dynamic_model_meshes()
        else:
            all_meshes = self.static_meshes + self.dynamic_meshes
        all_lines = self.static_lines + self.dynamic_lines
        if recenter:
            if (len(all_meshes)+len(all_lines)) == 0:
                return
            for m in all_meshes:
                m.v = m.v.reshape((-1, 3))
            all_verts = np.concatenate(
                [m.v[m.f.flatten() if m.f is not None and len(m.f) > 0 else np.arange(len(m.v))] for m in all_meshes] +
                [l.v[l.e.flatten()] for l in all_lines],
                axis=0)
            maximum = np.max(all_verts, axis=0)
            minimum = np.min(all_verts, axis=0)
            center = (maximum + minimum) / 2.
            scalefactor = (maximum - minimum) / 4.
#            scalefactor = (maximum - minimum) / 6.
            scalefactor = np.max(scalefactor)
        else:
            center = np.array(center)
        self.current_center = center
        self.current_scalefactor = scalefactor
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPushMatrix()
        # uncomment to add a default rotation (useful when automatically snapshoting kinect data
        #glRotate(220, 0.0, 1.0, 0.0)
        tf = np.identity(4, 'f')/scalefactor
        tf[:3, 3] = -center/scalefactor
        tf[3, 3] = 1
        cur_mtx = gl.glGetFloatv(gl.GL_MODELVIEW_MATRIX).T
        gl.glLoadMatrixf(cur_mtx.dot(tf).T)
        if want_camera:
            result = {
                'modelview_matrix': gl.glGetDoublev(gl.GL_MODELVIEW_MATRIX),
                'projection_matrix': gl.glGetDoublev(gl.GL_PROJECTION_MATRIX),
                'viewport': gl.glGetIntegerv(gl.GL_VIEWPORT)
            }
        else:
            result = None
        for m in all_meshes:
            if not hasattr(m, 'vbo'):
                # Precompute vertex vbo
                fidxs = m.f.flatten() if m.f is not None and len(m.f) > 0 else np.arange(len(m.v))
                allpts = m.v[fidxs].astype(np.float32).flatten()
                vbo = OpenGL.arrays.vbo.VBO(allpts)
                m.vbo = {'v': vbo}
                # Precompute normals vbo
                if m.vn is not None:
                    ns = m.vn.astype(np.float32)
                    ns = ns[m.f.flatten(), :]
                    m.vbo['vn'] = OpenGL.arrays.vbo.VBO(ns.flatten())
                elif m.f is not None and m.f.size > 0:
                    ns = TriNormals(m.v, m.f).r.reshape(-1, 3)
                    ns = np.tile(ns, (1, 3)).reshape(-1, 3).astype(np.float32)
                    m.vbo['vn'] = OpenGL.arrays.vbo.VBO(ns.flatten())
                # Precompute texture vbo
                if m.ft is not None and (m.ft.size > 0):
                    ftidxs = m.ft.flatten()
                    data = m.vt[ftidxs].astype(np.float32)[:, 0:2]
                    data[:, 1] = 1.0 - 1.0*data[:, 1]
                    m.vbo['vt'] = OpenGL.arrays.vbo.VBO(data)
                # Precompute color vbo
                if m.vc is not None:
                    data = m.vc[fidxs].astype(np.float32)
                    m.vbo['vc'] = OpenGL.arrays.vbo.VBO(data)
                elif m.fc is not None:
                    data = np.tile(m.fc, (1, 3)).reshape(-1, 3).astype(np.float32)
                    m.vbo['vc'] = OpenGL.arrays.vbo.VBO(data)
        for e in all_lines:
            self.draw_lines(e)
        for m in all_meshes:
            if m.texture_image is not None and not hasattr(m, 'textureID'):
                self.set_texture(m)
            self.draw_mesh(m)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPopMatrix()
        return result


class MeshViewerRemote(object):
    def __init__(self, titlebar='Mesh Viewer', subwins_vert=1, subwins_horz=1, width=100, height=100):
        context = zmq.Context.instance()
        self.server = context.socket(zmq.PULL)
        self.server.linger = 0
        #width = 1280
        #height = 960
        #width = 640
        #height = 480
        # Find a port to use. The standard set of "private" ports is 49152 through 65535, as seen in...
        # http://en.wikipedia.org/wiki/Port_(computer_networking)
        port = self.server.bind_to_random_port('tcp://127.0.0.1', min_port=49152, max_port=65535, max_tries=100000)
        # Print out our port so that our client can connect to us with it. Flush stdout immediately; otherwise
        # our client could wait forever.
        print '<PORT>%d</PORT>\n' % (port, )
        sys.stdout.flush()
        self.arcball = arcball.ArcBallT(width, height)
        self.transform = arcball.Matrix4fT()
        self.lastrot = arcball.Matrix3fT()
        self.thisrot = arcball.Matrix3fT()
        self.isdragging = False
        self.need_redraw = True
        self.mesh_viewers = [[MeshViewerSingle(float(c)/(subwins_horz), float(r)/(subwins_vert), 1./subwins_horz, 1./subwins_vert) for c in range(subwins_horz)] for r in range(subwins_vert)]
        self.tm_for_fps = 0.
        self.titlebar = titlebar
        self.activate(width, height)

    def snapshot(self, path):
        import cv
        self.on_draw()
        if False: # pylint: disable=using-constant-test
            x, y, width, height = gl.glGetIntegerv(gl.GL_VIEWPORT)
        else:
            x = 0
            y = 0
            width = glut.glutGet(glut.GLUT_WINDOW_WIDTH)
            height = glut.glutGet(glut.GLUT_WINDOW_HEIGHT)
        data = gl.glReadPixels(x, y, width, height, gl.GL_BGR, gl.GL_UNSIGNED_BYTE)
        image = cv.CreateImage((width, height), cv.IPL_DEPTH_8U, 3)
        flipped_image = cv.CreateImage((width, height), cv.IPL_DEPTH_8U, 3)
        cv.SetData(image, data)
        cv.Flip(image, flipped_image, 0)
        cv.SaveImage(path, flipped_image)

    def activate(self, width, height):
        glut.glutInit(['mesh_viewer'])
        glut.glutInitDisplayMode(glut.GLUT_RGBA | glut.GLUT_DOUBLE | glut.GLUT_ALPHA | glut.GLUT_DEPTH)
        glut.glutInitWindowSize(width, height)
        glut.glutInitWindowPosition(0, 0)
        self.root_window_id = glut.glutCreateWindow(self.titlebar)
        glut.glutDisplayFunc(self.on_draw)
        #glut.glutIdleFunc(self.checkQueue)
        glut.glutTimerFunc(100, self.checkQueue, 0)
        glut.glutReshapeFunc(self.on_resize_window)
        glut.glutKeyboardFunc(self.on_keypress)
        glut.glutMouseFunc(self.on_click)
        glut.glutMotionFunc(self.on_drag)
        # for r, lst in enumerate(self.mesh_viewers):
        #     for c, mv in enumerate(lst):
        #         mv.glut_window_id = glut.glutCreateSubWindow(self.root_window_id, c*width/len(lst), r*height/len(self.mesh_viewers), width/len(lst), height/len(self.mesh_viewers))
        glut.glutDisplayFunc(self.on_draw)
        self.init_opengl()
        glut.glutMainLoop() # won't return until process is killed

    def on_drag(self, cursor_x, cursor_y):
        """ Mouse cursor is moving
            Glut calls this function (when mouse button is down)
            and pases the mouse cursor postion in window coords as the mouse moves.
        """
        from blmath.geometry.transform.rodrigues import as_rotation_matrix
        if self.isdragging:
            mouse_pt = arcball.Point2fT(cursor_x, cursor_y)
            ThisQuat = self.arcball.drag(mouse_pt)                        # // Update End Vector And Get Rotation As Quaternion
            self.thisrot = arcball.Matrix3fSetRotationFromQuat4f(ThisQuat)        # // Convert Quaternion Into Matrix3fT
            # Use correct Linear Algebra matrix multiplication C = A * B
            self.thisrot = arcball.Matrix3fMulMatrix3f(self.lastrot, self.thisrot)      # // Accumulate Last Rotation Into This One
            # make sure it is a rotation
            self.thisrot = as_rotation_matrix(self.thisrot)
            self.transform = arcball.Matrix4fSetRotationFromMatrix3f(self.transform, self.thisrot)  # // Set Our Final Transform's Rotation From This One
            glut.glutPostRedisplay()
        return

    # The function called whenever a key is pressed. Note the use of Python tuples to pass in: (key, x, y)
    def on_keypress(self, *args):
        key = args[0]
        if hasattr(self, 'event_port'):
            self.keypress_port = self.event_port
            del self.event_port
        if hasattr(self, 'keypress_port'):
            client = zmq.Context.instance().socket(zmq.PUSH)
            client.connect('tcp://127.0.0.1:%d' % (self.keypress_port))
            client.send_pyobj({'event_type': 'keyboard', 'key': key})
            del self.keypress_port

    def on_click(self, button, button_state, cursor_x, cursor_y):
        """ Mouse button clicked.
            Glut calls this function when a mouse button is
            clicked or released.
        """
        self.isdragging = False
        # if (button == glut.GLUT_RIGHT_BUTTON and button_state == glut.GLUT_UP):
        #     # Right button click
        #     self.lastrot = Matrix3fSetIdentity()                         # // Reset Rotation
        #     self.thisrot = Matrix3fSetIdentity()                         # // Reset Rotation
        #     self.transform = Matrix4fSetRotationFromMatrix3f(self.transform, self.thisrot) # // Reset Rotation
        if button == glut.GLUT_LEFT_BUTTON and button_state == glut.GLUT_UP:
            # Left button released
            self.lastrot = copy.copy(self.thisrot)                          # // Set Last Static Rotation To Last Dynamic One
        elif button == glut.GLUT_LEFT_BUTTON and button_state == glut.GLUT_DOWN:
            # Left button clicked down
            self.lastrot = copy.copy(self.thisrot)                          # // Set Last Static Rotation To Last Dynamic One
            self.isdragging = True                                         # // Prepare For Dragging
            mouse_pt = arcball.Point2fT(cursor_x, cursor_y)
            self.arcball.click(mouse_pt)                             # // Update Start Vector And Prepare For Dragging
        elif button == glut.GLUT_RIGHT_BUTTON and button_state == glut.GLUT_DOWN:
            # If a mouse click location was requested, return it to caller
            if hasattr(self, 'event_port'):
                self.mouseclick_port = self.event_port
                del self.event_port
            if hasattr(self, 'mouseclick_port'):
                self.send_mouseclick_to_caller(cursor_x, cursor_y)
        elif button == glut.GLUT_MIDDLE_BUTTON and button_state == glut.GLUT_DOWN:
            # If a mouse click location was requested, return it to caller
            if hasattr(self, 'event_port'):
                self.mouseclick_port = self.event_port
                del self.event_port
            if hasattr(self, 'mouseclick_port'):
                self.send_mouseclick_to_caller(cursor_x, cursor_y, button='middle')
        glut.glutPostRedisplay()

    def send_mouseclick_to_caller(self, cursor_x, cursor_y, button='right'):
        client = zmq.Context.instance().socket(zmq.PUSH)
        client.connect('tcp://127.0.0.1:%d' % (self.mouseclick_port))
        cameras = self.on_draw(want_cameras=True)
        window_height = glut.glutGet(glut.GLUT_WINDOW_HEIGHT)
        depth_value = gl.glReadPixels(cursor_x, window_height-cursor_y, 1, 1, gl.GL_DEPTH_COMPONENT, gl.GL_FLOAT)
        pyobj = {
            'event_type': 'mouse_click_%sbutton' % button,
            'u': None, 'v': None,
            'x': None, 'y': None, 'z': None,
            'subwindow_row': None,
            'subwindow_col': None
            }
        for subwin_row, camera_list in enumerate(cameras):
            for subwin_col, camera in enumerate(camera_list):
                # test for out-of-bounds
                if cursor_x < camera['viewport'][0]:
                    continue
                if cursor_x > (camera['viewport'][0] + camera['viewport'][2]):
                    continue
                if window_height-cursor_y < camera['viewport'][1]:
                    continue
                if window_height-cursor_y > (camera['viewport'][1] + camera['viewport'][3]):
                    continue
                xx, yy, zz = glu.gluUnProject(
                    cursor_x, window_height-cursor_y, depth_value,
                    camera['modelview_matrix'],
                    camera['projection_matrix'],
                    camera['viewport'])
                pyobj = {
                    'event_type': 'mouse_click_%sbutton' % button,
                    'u': cursor_x - camera['viewport'][0], 'v': window_height - cursor_y - camera['viewport'][1],
                    'x': xx, 'y': yy, 'z': zz,
                    'which_subwindow': (subwin_row, subwin_col)
                    }
        client.send_pyobj(pyobj)
        del self.mouseclick_port

    def on_draw(self, want_cameras=False):
        #sys.stderr.write('fps: %.2e\n' % (1. / (time.time() - self.tm_for_fps)))
        self.tm_for_fps = time.time()
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        cameras = []
        for mvl in self.mesh_viewers:
            cameras.append([])
            for mv in mvl:
                cameras[-1].append(mv.on_draw(self.transform, want_cameras))
        gl.glFlush()                                                     # // Flush The GL Rendering Pipeline
        glut.glutSwapBuffers()
        self.need_redraw = False
        if want_cameras:
            return cameras

    # Reshape The Window When It's Moved Or Resized
    def on_resize_window(self, Width, Height):
        self.arcball.setBounds(Width, Height) # //*NEW* Update mouse bounds for arcball
        return

    def handle_request(self, request):
        label = request['label']
        obj = request['obj']
        w = request['which_window']
        mv = self.mesh_viewers[w[0]][w[1]]
        # Handle each type of request.
        # Some requests require a redraw, and
        # some don't.
        if label == 'dynamic_meshes':
            mv.dynamic_meshes = obj
            self.need_redraw = True
        elif label == 'dynamic_models':
            mv.dynamic_models = obj
            self.need_redraw = True
        elif label == 'static_meshes':
            mv.static_meshes = obj
            self.need_redraw = True
        elif label == 'dynamic_lines':
            mv.dynamic_lines = obj
            self.need_redraw = True
        elif label == 'static_lines':
            mv.static_lines = obj
            self.need_redraw = True
        elif label == 'autorecenter':
            mv.autorecenter = obj
            self.need_redraw = True
        elif label == 'titlebar':
            assert isinstance(obj, str) or isinstance(obj, unicode)
            self.titlebar = obj
            glut.glutSetWindowTitle(obj)
        elif label == 'background_color':
            gl.glClearColor(obj[0], obj[1], obj[2], 1.0)
            self.need_redraw = True
        elif label == 'save_snapshot': # redraws for itself
            assert isinstance(obj, str) or isinstance(obj, unicode)
            self.snapshot(obj)
        elif label == 'get_keypress':
            self.keypress_port = obj
        elif label == 'get_mouseclick':
            self.mouseclick_port = obj
        elif label == 'get_event':
            self.event_port = obj
        else:
            return False # can't handle this request string
        return True # handled the request string

    def checkQueue(self, unused_timer_id): # pylint: disable=unused-argument
        glut.glutTimerFunc(20, self.checkQueue, 0)
        # if True: # spinning
        #     w_whole_window = glutGet(GLUT_WINDOW_WIDTH)
        #     h_whole_window = glutGet(GLUT_WINDOW_HEIGHT)
        #     center_x = w_whole_window/2
        #     center_y = h_whole_window/2
        #     self.on_click(GLUT_LEFT_BUTTON, GLUT_DOWN, center_x, center_y)
        #     self.on_drag(center_x+2, center_y)
        try:
            request = self.server.recv_pyobj(zmq.NOBLOCK)
        except zmq.ZMQError as e:
            if e.errno != zmq.EAGAIN:
                raise # something wrong besides empty queue
            return # empty queue, no problem
        if not request:
            return
        while request:
            task_completion_time = time.time()
            if not self.handle_request(request):
                raise Exception('Unknown command string: %s' % (request['label']))
            task_completion_time = time.time() - task_completion_time
            if 'port' in request: # caller wants confirmation
                port = request['port']
                client = zmq.Context.instance().socket(zmq.PUSH)
                client.connect('tcp://127.0.0.1:%d' % (port))
                client.send_pyobj(task_completion_time)
            try:
                request = self.server.recv_pyobj(zmq.NOBLOCK)
            except zmq.ZMQError as e:
                if e.errno != zmq.EAGAIN:
                    raise
                request = None
        if self.need_redraw:
            glut.glutPostRedisplay()

    # A general OpenGL initialization function.  Sets all of the initial parameters.
    def init_opengl(self):             # We call this right after our OpenGL window is created.
        gl.glClearColor(0.0, 0.0, 0.0, 1.0)                    # This Will Clear The Background Color To Black
        gl.glClearDepth(1.0)                                   # Enables Clearing Of The Depth Buffer
        gl.glDepthFunc(gl.GL_LEQUAL)                              # The Type Of Depth Test To Do
        gl.glEnable(gl.GL_DEPTH_TEST)                             # Enables Depth Testing
        gl.glShadeModel(gl.GL_SMOOTH)
        gl.glHint(gl.GL_PERSPECTIVE_CORRECTION_HINT, gl.GL_NICEST)  # Really Nice Perspective Calculations
        gl.glEnable(gl.GL_LIGHT0)
        gl.glEnable(gl.GL_LIGHTING)
        gl.glEnable(gl.GL_COLOR_MATERIAL)
        gl.glEnable(gl.GL_NORMALIZE) # important since we rescale the modelview matrix
        # gl.glPolygonMode( gl.GL_FRONT_AND_BACK, gl.GL_LINE )
        return True


def main(argv):
    if len(argv) > 2:
        # Start a remote window
        MeshViewerRemote(titlebar=argv[1], subwins_vert=int(argv[2]), subwins_horz=int(argv[3]), width=int(argv[4]), height=int(argv[5]))
    elif len(argv) == 2:
        # Just a test for opengl working
        try:
            from OpenGL.GLUT import glutInit
            glutInit()
        except Exception as e: # pylint: disable=broad-except
            print >>sys.stderr, e
            print 'failure'
        else:
            print 'success'

if __name__ == '__main__':
    freeze_support() # Windows specific: see http://docs.python.org/2/library/multiprocessing.html#multiprocessing.freeze_support
    main(sys.argv)



def load(f, existing_mesh=None):
    from baiji.serialization.util.openlib import ensure_file_open_and_call
    return ensure_file_open_and_call(f, _load, mode='rb', existing_mesh=existing_mesh)

def _load(f, existing_mesh=None):
    import os
    from lace.serialization import ply, obj, wrl, stl, xyz, bsf
    ext = os.path.splitext(f.name)[1].lower()
    if ext == ".ply":
        return ply.load(f, existing_mesh=existing_mesh)
    elif ext == ".obj":
        return obj.load(f, existing_mesh=existing_mesh)
    elif ext == ".wrl":
        return wrl.load(f, existing_mesh=existing_mesh)
    elif ext == ".ylp":
        return ply.load_encrypted(f, existing_mesh=existing_mesh)
    elif ext == ".stl":
        return stl.load(f, existing_mesh=existing_mesh)
    elif ext == ".xyz":
        return xyz.load(f, existing_mesh=existing_mesh)
    elif ext == ".bsf":
        return bsf.load(f, existing_mesh=existing_mesh)
    else:
        raise NotImplementedError("Unknown mesh file format: %s" % f.name)



# NB: setter_property incorrectly trigger's pylint's method-hidden error
# pylint: disable=method-hidden, unused-argument, redefined-outer-name
import numpy as np
from blmath.numerics import as_numeric_array
from blmath.util.decorators import setter_property

from lace import serialization
from lace import color
from lace import geometry
from lace import landmarks
from lace import topology
from lace import visibility
from lace import texture
from lace import search
from lace import visualization

class Mesh(
        serialization.MeshMixin,
        color.MeshMixin,
        geometry.MeshMixin,
        landmarks.MeshMixin,
        topology.MeshMixin,
        visibility.MeshMixin,
        texture.MeshMixin,
        search.MeshMixin,
        visualization.MeshMixin,
        object
    ):
    """3d Triangulated Mesh class

    Attributes:
        v: Vx3 array of vertices
        f: Fx3 array of faces
        f4: Fx4 array of faces (optional quads)
        e: E*2 array of edges

    Optional attributes:
        fc: Fx3 array of face colors
        vc: Vx3 array of vertex colors
        vn: Vx3 array of vertex normals
        segm: dictionary of part names to triangle indices

    """
    def __init__(self,
                 filename=None,
                 v=None, f=None, f4=None,
                 vc=None, fc=None,
                 vn=None, fn=None,
                 vt=None, ft=None,
                 e=None, segm=None,
                 basename=None,
                 landmarks=None, ppfilename=None, lmrkfilename=None):
        _properties = ['v', 'f', 'f4', 'fc', 'vc', 'fn', 'vn', 'ft', 'vt', 'e', 'segm']
        # First, set initial values, so that unset v doesn't override the v
        # from a file. We do this explicitly rather than using _properties
        # so that pylint will recognize the attributes and make
        # attribute-defined-outside-init warnings useful.
        self.v = self.f = self.f4 = self.fc = self.vc = self.vn = self.fn = self.vt = self.ft = self.e = self.segm = None
        self.basename = basename

        # Load from a file, if one is specified
        if filename is not None:
            if isinstance(filename, basestring):
                # Support `Mesh(filename)`
                from lace.serialization import mesh
                mesh.load(filename, existing_mesh=self)
                if self.basename is None:
                    import os
                    self.basename = os.path.splitext(os.path.basename(filename))[0]
                self.filename = filename
            elif isinstance(filename, Mesh) or hasattr(filename, 'v'):
                # Support `Mesh(mesh)`
                import copy
                other_mesh = filename
                # A deep copy with all of the numpy arrays copied:
                for a in ['v', 'f'] + other_mesh.__dict__.keys(): # NB: v and f first as vc and fc need them
                    if a == 'landm_raw_xyz':
                        # We've deprecated landm_raw_xyz and it raises an error to access it now, but some
                        # older pickled meshes (don't pickle meshes!) still have it as a property and they
                        # raise an error on the default getattr call (in addition to not converting the
                        # data to the current format), so we special case this.
                        self.landm_xyz = other_mesh.__dict__['landm_raw_xyz']
                    else:
                        setattr(self, a, copy.deepcopy(getattr(other_mesh, a)))
            elif v is None:
                # Historically, meshes were crated with Mesh(verts)
                try:
                    v = np.array(filename, dtype=np.float64)
                except ValueError:
                    pass

        # And then override whatever came from the file or the other
        # mesh with explicitly given values
        for a in _properties:
            if locals()[a] is not None:
                setattr(self, a, locals()[a])

        # The following should probably be cleaned up at some point...
        if landmarks is not None:
            self.landm = landmarks
        if ppfilename is not None:
            self.landm = ppfilename
        if lmrkfilename is not None:
            self.landm = lmrkfilename

    def __del__(self):
        if hasattr(self, 'textureID'):
            # This may be set by MeshViewer
            from OpenGL.GL import glDeleteTextures
            glDeleteTextures([self.textureID])

    # These control pickling and unpickling:
    # We need this extra little tweak to handle loading pickled meshes that have member variables
    # that have since been changed to @properties. Without this, the member just doesn't get loaded
    # from the pickle.
    def __getstate__(self):
        # Same as the default implementation, but if it's not defined then __setstate__ is not called
        return self.__dict__

    def __setstate__(self, d):
        MEMBERS_CHANGED_TO_PROPERTIES = ['texture_filepath']
        for name in MEMBERS_CHANGED_TO_PROPERTIES:
            if name in d:
                d['_' + name] = d[name]
                del d[name]
        self.__dict__ = d # come on pylint, __dict__ is defined *before* __init__ pylint: disable=attribute-defined-outside-init

    def copy(self, only=None):
        '''
        Returns a deep copy with all of the numpy arrays copied
        If only is a list of strings, i.e. ['f', 'v'], then only those properties will be copied
        '''
        if only is None:
            return Mesh(self)
        else:
            import copy
            m = Mesh()
            for a in only:
                setattr(m, a, copy.deepcopy(getattr(self, a)))
            return m

    def copy_fv(self):
        '''Sugar on top of copy to only copy the basic f & v attributes'''
        return self.copy(only=['f', 'v'])

    def copy_fvs(self):
        '''Sugar on top of copy to only copy the basic f, v, & segm
        attributes
        '''
        return self.copy(only=['f', 'v', 'segm'])

    # Rather than use a private _x varaible and boilerplate getters for
    # these, we'll use the actual var name and just override the setter.
    @setter_property
    def v(self, val):
        # cached properties that are dependent on v
        self._clear_cached_properties('vertices_to_edges_matrix', 'vertices_to_edges_matrix_single_axis')
        self.__dict__['v'] = as_numeric_array(val, dtype=np.float64, shape=(-1, 3), allow_none=True, empty_as_none=True)

    @setter_property
    def f(self, val):
        # cached properties that are dependent on f
        self._clear_cached_properties('faces_per_edge', 'vertices_per_edge', 'vertices_to_edges_matrix', 'vertices_to_edges_matrix_single_axis')
        self.__dict__['f'] = as_numeric_array(val, dtype=np.uint64, shape=(-1, 3), allow_none=True, empty_as_none=True)

    @setter_property
    def f4(self, val):
        self.__dict__['f4'] = as_numeric_array(val, dtype=np.uint64, shape=(-1, 4), allow_none=True, empty_as_none=True)

    @setter_property
    def fc(self, val):
        val = color.colors_like(val, self.f)
        self.__dict__['fc'] = val

    @setter_property
    def vc(self, val):
        val = color.colors_like(val, self.v)
        self.__dict__['vc'] = val

    @setter_property
    def fn(self, val):
        '''
        Note that in some applications, face normals are vectors, in others they are indexes into the vertex normal array.
        Someday we should refactor things so that we have seperate fn (face normal vectors) & fvn (face vertex normal indicies).
        Today is not that day.
        '''
        self.__dict__['fn'] = as_numeric_array(val, dtype=np.float64, shape=(-1, 3), allow_none=True, empty_as_none=True)

    @setter_property
    def vn(self, val):
        self.__dict__['vn'] = as_numeric_array(val, dtype=np.float64, shape=(-1, 3), allow_none=True, empty_as_none=True)

    @setter_property
    def ft(self, val):
        self.__dict__['ft'] = as_numeric_array(val, dtype=np.uint32, shape=(-1, 3), allow_none=True, empty_as_none=True)

    @setter_property
    def vt(self, val):
        self.__dict__['vt'] = as_numeric_array(val, dtype=np.float64, shape=(-1, 2), allow_none=True, empty_as_none=True)

    @setter_property
    def e(self, val):
        self.__dict__['e'] = as_numeric_array(val, dtype=np.uint64, shape=(-1, 2), allow_none=True, empty_as_none=True)

    @setter_property
    def segm(self, val):
        self.__dict__['segm'] = val

    def _clear_cached_properties(self, *keys):
        for cached_property_key in keys:
            if cached_property_key in self.__dict__:
                del self.__dict__[cached_property_key]

    @classmethod
    def concatenate(cls, *args):
        """Concatenates an arbitrary number of meshes.

        Currently supports vertices, vertex colors, and faces.
        """
        nargs = len(args)
        if nargs == 1:
            return args[0]

        vs = [a.v for a in args if a.v is not None]
        vcs = [a.vc for a in args if a.vc is not None]
        fs = [a.f for a in args if a.f is not None]

        if vs and len(vs) != nargs:
            raise ValueError('Expected `v` for all args or none.')
        if vcs and len(vcs) != nargs:
            raise ValueError('Expected `vc` for all args or none.')
        if fs and len(fs) != nargs:
            raise ValueError('Expected `f` for all args or none.')

        # Offset face indices by the cumulative vertex count.
        face_offsets = np.cumsum([v.shape[0] for v in vs[:-1]])
        for offset, f in zip(face_offsets, fs[1:]):  # https://bitbucket.org/logilab/pylint/issues/603/operator-generates-false-positive-unused pylint: disable=unused-variable
            f += offset

        return Mesh(
            v=np.vstack(vs) if vs else None,
            vc=np.vstack(vcs) if vcs else None,
            f=np.vstack(fs) if fs else None,
        )



class MeshMixin(object):
    '''
    Convenience methods for serialization.
    Please write serializers rather than adding explicit serialization code in here.
    '''

    def write(self, filename):
        import os
        _, fileext = os.path.splitext(filename)
        if fileext == '.obj':
            self.write_obj(filename)
        elif fileext == '.ply':
            self.write_ply(filename)
        elif fileext == '.dae':
            self.write_dae(filename)
        else:
            raise TypeError('Unsupported filetype %s' % fileext)

    def write_obj(self, filename, flip_faces=False, ungroup=False, comments=None, write_mtl=True):
        from lace.serialization import obj
        obj.dump(self, filename, flip_faces=flip_faces, ungroup=ungroup, comments=comments, write_mtl=write_mtl)

    def write_fuse_obj(self, filename):
        from lace.serialization import obj
        obj.dump(self, filename, split_normals=True)

    def write_mtl(self, path, material_name, texture_name):
        from lace.serialization import obj
        obj.write_mtl(path, material_name, texture_name)

    def write_ply(self, filename, flip_faces=False, ascii=False, little_endian=True, comments=[]):
        from lace.serialization import ply
        ply.dump(self, filename, flip_faces=flip_faces, ascii=ascii, little_endian=little_endian, comments=comments)

    def write_dae(self, filename):
        from lace.serialization import dae
        dae.dump(self, filename)



"""
ArcBall.py -- Math utilities, vector, matrix types and ArcBall quaternion rotation class
"""
import copy
from math import sqrt
import numpy as np

# //assuming IEEE-754(GLfloat), which i believe has max precision of 7 bits
Epsilon = 1.0e-5


class ArcBallT(object):
    def __init__(self, NewWidth, NewHeight):
        self.m_StVec = Vector3fT()
        self.m_EnVec = Vector3fT()
        self.m_AdjustWidth = 1.0
        self.m_AdjustHeight = 1.0
        self.setBounds(NewWidth, NewHeight)

    def __str__(self):
        str_rep = ""
        str_rep += "StVec = " + str(self.m_StVec)
        str_rep += "\nEnVec = " + str(self.m_EnVec)
        str_rep += "\n scale coords %f %f" % (self.m_AdjustWidth, self.m_AdjustHeight)
        return str_rep

    def setBounds(self, NewWidth, NewHeight):
        # //Set new bounds
        assert (NewWidth > 1.0 and NewHeight > 1.0), "Invalid width or height for bounds."
        # //Set adjustment factor for width/height
        self.m_AdjustWidth = 1.0 / ((NewWidth - 1.0) * 0.5)
        self.m_AdjustHeight = 1.0 / ((NewHeight - 1.0) * 0.5)

    def _mapToSphere(self, NewPt):
        # Given a new window coordinate, will modify NewVec in place
        X = 0
        Y = 1
        Z = 2

        NewVec = Vector3fT()
        # //Copy paramter into temp point
        TempPt = copy.copy(NewPt)
        # //Adjust point coords and scale down to range of [-1 ... 1]
        TempPt[X] = (NewPt[X] * self.m_AdjustWidth) - 1.0
        TempPt[Y] = 1.0 - (NewPt[Y] * self.m_AdjustHeight)
        # //Compute the square of the length of the vector to the point from the center
        length = np.sum(np.dot(TempPt, TempPt))
        # //If the point is mapped outside of the sphere... (length > radius squared)
        if length > 1.0:
            # //Compute a normalizing factor (radius / sqrt(length))
            norm = 1.0 / sqrt(length)

            # //Return the "normalized" vector, a point on the sphere
            NewVec[X] = TempPt[X] * norm
            NewVec[Y] = TempPt[Y] * norm
            NewVec[Z] = 0.0
        else:           # //Else it's on the inside
            # //Return a vector to a point mapped inside the sphere sqrt(radius squared - length)
            NewVec[X] = TempPt[X]
            NewVec[Y] = TempPt[Y]
            NewVec[Z] = sqrt(1.0 - length)

        return NewVec

    def click(self, NewPt):
        # //Mouse down (Point2fT
        self.m_StVec = self._mapToSphere(NewPt)
        return

    def drag(self, NewPt):
        # //Mouse drag, calculate rotation (Point2fT Quat4fT)
        """ drag (Point2fT mouse_coord) -> new_quaternion_rotation_vec
        """
        X = 0
        Y = 1
        Z = 2
        W = 3

        self.m_EnVec = self._mapToSphere(NewPt)

        # //Compute the vector perpendicular to the begin and end vectors
        # Perp = Vector3fT()
        Perp = Vector3fCross(self.m_StVec, self.m_EnVec)

        NewRot = Quat4fT()
        # //Compute the length of the perpendicular vector
        if Vector3fLength(Perp) > Epsilon:        #    //if its non-zero
            # //We're ok, so return the perpendicular vector as the transform after all
            NewRot[X] = Perp[X]
            NewRot[Y] = Perp[Y]
            NewRot[Z] = Perp[Z]
            # //In the quaternion values, w is cosine(theta / 2), where theta is rotation angle
            NewRot[W] = Vector3fDot(self.m_StVec, self.m_EnVec)
        else:       #                            //if its zero
            # //The begin and end vectors coincide, so return a quaternion of zero matrix (no rotation)
            NewRot[X] = NewRot[Y] = NewRot[Z] = NewRot[W] = 0.0

        return NewRot


# ##################### Math utility ##########################################


def Matrix4fT():
    return np.identity(4, 'f')

def Matrix3fT():
    return np.identity(3, 'f')

def Quat4fT():
    return np.zeros(4, 'f')

def Vector3fT():
    return np.zeros(3, 'f')

def Point2fT(x=0.0, y=0.0):
    pt = np.zeros(2, 'f')
    pt[0] = x
    pt[1] = y
    return pt

def Vector3fDot(u, v):
    # Dot product of two 3f vectors
    dotprod = np.dot(u, v)
    return dotprod

def Vector3fCross(u, v):
    # Cross product of two 3f vectors
    X = 0
    Y = 1
    Z = 2
    cross = np.zeros(3, 'f')
    cross[X] = (u[Y] * v[Z]) - (u[Z] * v[Y])
    cross[Y] = (u[Z] * v[X]) - (u[X] * v[Z])
    cross[Z] = (u[X] * v[Y]) - (u[Y] * v[X])
    return cross

def Vector3fLength(u):
    mag_squared = np.sum(np.dot(u, u))
    mag = sqrt(mag_squared)
    return mag

def Matrix3fSetIdentity():
    return np.identity(3, 'f')

def Matrix3fMulMatrix3f(matrix_a, matrix_b):
    return matrix_a.dot(matrix_b)




def Matrix4fSVD(NewObj):
    X = 0
    Y = 1
    Z = 2
    s = sqrt(
        ((NewObj[X][X] * NewObj[X][X]) + (NewObj[X][Y] * NewObj[X][Y]) + (NewObj[X][Z] * NewObj[X][Z]) +
         (NewObj[Y][X] * NewObj[Y][X]) + (NewObj[Y][Y] * NewObj[Y][Y]) + (NewObj[Y][Z] * NewObj[Y][Z]) +
         (NewObj[Z][X] * NewObj[Z][X]) + (NewObj[Z][Y] * NewObj[Z][Y]) + (NewObj[Z][Z] * NewObj[Z][Z])) / 3.0)
    return s

def Matrix4fSetRotationScaleFromMatrix3f(NewObj, three_by_three_matrix):
    # Modifies NewObj in-place by replacing its upper 3x3 portion from the
    # passed in 3x3 matrix.
    # NewObj = Matrix4fT ()
    NewObj[0:3, 0:3] = three_by_three_matrix
    return NewObj

# /**
# * Sets the rotational component (upper 3x3) of this matrix to the matrix
# * values in the T precision Matrix3d argument; the other elements of
# * this matrix are unchanged; a singular value decomposition is performed
# * on this object's upper 3x3 matrix to factor out the scale, then this
# * object's upper 3x3 matrix components are replaced by the passed rotation
# * components, and then the scale is reapplied to the rotational
# * components.
# * @param three_by_three_matrix T precision 3x3 matrix
# */
def Matrix4fSetRotationFromMatrix3f(NewObj, three_by_three_matrix):
    scale = Matrix4fSVD(NewObj)

    NewObj = Matrix4fSetRotationScaleFromMatrix3f(NewObj, three_by_three_matrix)
    scaled_NewObj = NewObj * scale           # Matrix4fMulRotationScale(NewObj, scale)
    return scaled_NewObj

def Matrix3fSetRotationFromQuat4f(q1):
    # Converts the H quaternion q1 into a new equivalent 3x3 rotation matrix.
    X = 0
    Y = 1
    Z = 2
    W = 3

    NewObj = Matrix3fT()
    n = np.sum(np.dot(q1, q1))
    s = 0.0
    if n > 0.0:
        s = 2.0 / n

    # pylint: disable=multiple-statements
    xs = q1[X] * s;  ys = q1[Y] * s;  zs = q1[Z] * s
    wx = q1[W] * xs; wy = q1[W] * ys; wz = q1[W] * zs
    xx = q1[X] * xs; xy = q1[X] * ys; xz = q1[X] * zs
    yy = q1[Y] * ys; yz = q1[Y] * zs; zz = q1[Z] * zs
    # This math all comes about by way of algebra, complex math, and trig identities.
    # See Lengyel pages 88-92
    NewObj[X][X] = 1.0 - (yy + zz);    NewObj[Y][X] = xy - wz;            NewObj[Z][X] = xz + wy
    NewObj[X][Y] = xy + wz;            NewObj[Y][Y] = 1.0 - (xx + zz);    NewObj[Z][Y] = yz - wx
    NewObj[X][Z] = xz - wy;            NewObj[Y][Z] = yz + wx;            NewObj[Z][Z] = 1.0 - (xx + yy)

    return NewObj
